# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Metaclasses used in various explicit classes in PyFluent."""

from abc import ABCMeta
from collections.abc import MutableMapping
import inspect
from pprint import pformat
from typing import List

from ansys.fluent.core.exceptions import DisallowedValuesError, InvalidArgument

# pylint: disable=unused-private-member
# pylint: disable=bad-mcs-classmethod-argument


class Attribute:
    """Attributes."""

    VALID_NAMES = [
        "range",
        "allowed_values",
        "display_name_allowed_values",
        "help_str",
        "is_read_only",
        "is_active",
        "on_create",
        "on_change",
        "show_as_separate_object",
        "display_text",
        "layout",
        "previous",
        "next",
        "include",
        "exclude",
        "sort_by",
        "style",
        "icon",
        "show_text",
        "widget",
        "dir_info",
        "extensions",
    ]

    def __init__(self, function):
        self.function = function

    def __set_name__(self, obj, name):
        if name not in self.VALID_NAMES:
            raise DisallowedValuesError("attribute", name, self.VALID_NAMES)
        self.name = name
        if not hasattr(obj, "attributes"):
            obj.attributes = set()
        obj.attributes.add(name)

    def __set__(self, obj, value):
        raise AttributeError("Attributes are read only.")

    def __get__(self, obj, objtype=None):
        return self.function(obj)


class Command:
    """Executes command."""

    def __init__(self, method):
        self.arguments_attrs = {}
        cmd_args = inspect.signature(method).parameters
        for arg_name in cmd_args:
            if arg_name != "self":
                self.arguments_attrs[arg_name] = {}

        def _init(_self, obj):
            _self.obj = obj

        def _execute(_self, *args, **kwargs):
            for arg, attr_data in self.arguments_attrs.items():
                arg_value = None
                if arg in kwargs:
                    arg_value = kwargs[arg]
                else:
                    index = list(self.arguments_attrs.keys()).index(arg)
                    if len(args) > index:
                        arg_value = args[index]
                if arg_value is not None:
                    for attr, attr_value in attr_data.items():
                        if attr == "allowed_values":
                            allowed_values = attr_value(_self.obj)
                            if isinstance(arg_value, list):
                                if not all(
                                    elem in allowed_values for elem in arg_value
                                ):
                                    raise DisallowedValuesError(
                                        arg, arg_value, allowed_values
                                    )
                            else:
                                if arg_value not in allowed_values:
                                    raise DisallowedValuesError(
                                        arg, arg_value, allowed_values
                                    )

                        elif attr == "range":
                            if not isinstance(arg_value, (int, float)):
                                raise RuntimeError(
                                    f"{arg} value {arg_value} is not number."
                                )

                            minimum, maximum = attr_value(_self.obj)
                            if arg_value < minimum or arg_value > maximum:
                                raise DisallowedValuesError(
                                    arg, arg_value, allowed_values
                                )
            return method(_self.obj, *args, **kwargs)

        self.command_cls = type(
            "command",
            (),
            {
                "__init__": _init,
                "__call__": _execute,
                "argument_attribute": lambda _self, argument_name, attr_name: self.arguments_attrs[
                    argument_name
                ][
                    attr_name
                ](
                    _self.obj
                ),
                "arguments": lambda _self: list(self.arguments_attrs.keys()),
            },
        )

    def __set_name__(self, obj, name):
        self.obj = obj
        if not hasattr(obj, "commands"):
            obj.commands = {}
        obj.commands[name] = {}

    def __get__(self, obj, obj_type=None):
        if hasattr(self, "command"):
            return self.command
        else:
            return self.command_cls(obj)


def CommandArgs(command_object, argument_name):
    """Command arguments."""

    def wrapper(attribute):
        if argument_name in command_object.arguments_attrs:
            command_object.arguments_attrs[argument_name].update(
                {attribute.__name__: attribute}
            )
        else:
            raise InvalidArgument(f"{argument_name} not a valid argument.")
        return attribute

    return wrapper


class PyLocalBaseMeta(type):
    """Local base metaclass."""

    @classmethod
    def __create_get_ancestors_by_type(cls):
        def wrapper(self, obj_type, obj=None):
            obj = self if obj is None else obj
            parent = None
            if getattr(obj, "_parent", None):
                if isinstance(obj._parent, obj_type):
                    return obj._parent
                parent = self.get_ancestors_by_type(obj_type, obj._parent)
            return parent

        return wrapper

    @classmethod
    def __create_get_ancestors_by_name(cls):
        def wrapper(self, obj_type, obj=None):
            obj = self if obj is None else obj
            parent = None
            if getattr(obj, "_parent", None):
                if obj._parent.__class__.__name__ == obj_type:
                    return obj._parent
                if getattr(obj._parent, "PLURAL", None) == obj_type:
                    return obj._parent._parent
                parent = self.get_ancestors_by_name(obj_type, obj._parent)
            return parent

        return wrapper

    @classmethod
    def __create_get_root(cls):
        def wrapper(self, obj=None):
            obj = self if obj is None else obj
            parent = obj
            if getattr(obj, "_parent", None):
                parent = self.get_root(obj._parent)
            return parent

        return wrapper

    @classmethod
    def __create_get_session(cls):
        def wrapper(self, obj=None):
            root = self.get_root(obj)
            return root.session

        return wrapper

    @classmethod
    def __create_get_session_handle(cls):
        def wrapper(self, obj=None):
            root = self.get_root(obj)
            return getattr(root, "session_handle", None)

        return wrapper

    @classmethod
    def __create_get_path(cls):
        def wrapper(self):
            if getattr(self, "_parent", None):
                return self._parent.get_path() + "/" + self._name
            return self._name

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["get_ancestors_by_type"] = cls.__create_get_ancestors_by_type()
        attrs["get_ancestors_by_name"] = cls.__create_get_ancestors_by_name()
        attrs["get_root"] = cls.__create_get_root()
        attrs["get_session"] = cls.__create_get_session()
        attrs["get_session_handle"] = cls.__create_get_session_handle()
        if "get_path" not in attrs:
            attrs["get_path"] = cls.__create_get_path()
        attrs["root"] = property(lambda self: self.get_root())
        attrs["path"] = property(lambda self: self.get_path())
        attrs["session"] = property(lambda self: self.get_session())
        attrs["session_handle"] = property(lambda self: self.get_session_handle())
        return super(PyLocalBaseMeta, cls).__new__(cls, name, bases, attrs)


class PyLocalPropertyMeta(PyLocalBaseMeta):
    """Metaclass for local property classes."""

    @classmethod
    def __create_validate(cls):
        def wrapper(self, value):
            attrs = getattr(self, "attributes", None)
            if attrs:
                for attr in attrs:
                    if attr == "range":
                        if self.range and (
                            value < self.range[0] or value > self.range[1]
                        ):
                            raise DisallowedValuesError("value", value, self.range)
                    elif attr == "allowed_values":
                        if isinstance(value, list):
                            if not all(
                                v is None or v in self.allowed_values for v in value
                            ):
                                raise DisallowedValuesError(
                                    "value", value, self.allowed_values
                                )
                        elif value is not None and value not in self.allowed_values:
                            raise DisallowedValuesError(
                                "value", value, self.allowed_values
                            )
            return value

        return wrapper

    @classmethod
    def __create_init(cls):
        def wrapper(self, parent, api_helper, name=""):
            """Create the initialization method for 'PyLocalPropertyMeta'."""
            self._name = name
            self._api_helper = api_helper(self)
            self._parent = parent
            self._on_change_cbs = []
            annotations = self.__class__.__dict__.get("__annotations__")
            if isinstance(getattr(self.__class__, "value", None), property):
                value_annotation = annotations.get("_value")
            else:
                value_annotation = annotations.get("value")
            self.type = value_annotation
            reset_on_change = (
                hasattr(self, "_reset_on_change")
                and getattr(self, "_reset_on_change")()
            )

            on_change = getattr(self, "on_change", None)
            if on_change is not None:
                self._register_on_change_cb(on_change)
            if reset_on_change:
                for obj in reset_on_change:

                    def reset():
                        setattr(self, "_value", None)
                        for on_change_cb in self._on_change_cbs:
                            on_change_cb()

                    obj._register_on_change_cb(reset)

        return wrapper

    @classmethod
    def __create_get_state(cls, show_attributes=False):
        def wrapper(self):
            rv = self.value

            if hasattr(self, "allowed_values"):
                allowed_values = self.allowed_values
                if len(allowed_values) > 0 and (
                    rv is None
                    or (not isinstance(rv, list) and rv not in allowed_values)
                ):
                    self.set_state(allowed_values[0])
                    rv = self.value

            return rv

        return wrapper

    @classmethod
    def __create_set_state(cls):
        def wrapper(self, value, validate=True):
            self.value = self._validate(value) if validate else value
            for on_change_cb in self._on_change_cbs:
                on_change_cb()

        return wrapper

    @classmethod
    def __create_register_on_change(cls):
        def wrapper(self, on_change_cb):
            self._on_change_cbs.append(on_change_cb)

        return wrapper

    @classmethod
    def __create_repr(cls):
        def wrapper(self):
            data = self()
            return f"{data}"

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["__init__"] = cls.__create_init()
        attrs["__call__"] = cls.__create_get_state()
        attrs["__repr__"] = cls.__create_repr()
        attrs["_validate"] = cls.__create_validate()
        attrs["_register_on_change_cb"] = cls.__create_register_on_change()
        attrs["set_state"] = cls.__create_set_state()
        return super(PyLocalPropertyMeta, cls).__new__(cls, name, bases, attrs)


class PyReferenceObjectMeta(PyLocalBaseMeta):
    """Metaclass for local object classes."""

    @classmethod
    def __create_init(cls):
        def wrapper(self, parent, path, location, session_id, name=""):
            """Create the initialization method for 'PyReferenceObjectMeta'."""
            self._parent = parent
            self.type = "object"
            self.parent = parent
            self._path = path
            self.location = location
            self.session_id = session_id

            def update(clss):
                for name, cls in clss.__dict__.items():
                    if cls.__class__.__name__ in (
                        "PyLocalPropertyMeta",
                        "PyLocalObjectMeta",
                    ):
                        setattr(
                            self,
                            name,
                            cls(self, lambda arg: None, name),
                        )
                    if (
                        cls.__class__.__name__ == "PyLocalNamedObjectMeta"
                        or cls.__class__.__name__ == "PyLocalNamedObjectMetaAbstract"
                    ):
                        setattr(
                            self,
                            cls.PLURAL,
                            PyLocalContainer(self, cls, lambda arg: None, cls.PLURAL),
                        )
                for base_class in clss.__bases__:
                    update(base_class)

            update(self.__class__)

        return wrapper

    @classmethod
    def __create_get_path(cls):
        def wrapper(self):
            return self._path

        return wrapper

    @classmethod
    def __create_reset(cls):
        def wrapper(self, path, location, session_id):
            self._path = path
            self.location = location
            self.session_id = session_id
            if hasattr(self, "_object"):
                delattr(self, "_object")

        return wrapper

    @classmethod
    def __create_getattr(cls):
        def wrapper(self, item):
            if item == "_object":
                top_most_parent = self.get_root(self)

                if self.session_id is None:
                    self.session_id = top_most_parent.session.id
                property_editor_data = top_most_parent.accessor(
                    "AnsysUser", self.session_id
                )
                (
                    obj,
                    cmd_data,
                ) = property_editor_data.get_object_and_command_data_from_properties_info(
                    {"path": self.path, "properties": {}, "type": self.location}
                )
                if obj is not None:
                    self._object = obj
                return obj
            if item == "ref":
                return self._object._object

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["__init__"] = attrs.get("__init__", cls.__create_init())
        attrs["__getattr__"] = attrs.get("__getattr__", cls.__create_getattr())
        attrs["reset"] = cls.__create_reset()
        attrs["get_path"] = cls.__create_get_path()
        return super(PyReferenceObjectMeta, cls).__new__(cls, name, bases, attrs)


class PyLocalObjectMeta(PyLocalBaseMeta):
    """Metaclass for local object classes."""

    @classmethod
    def __create_init(cls):
        def wrapper(self, parent, api_helper, name=""):
            """Create the initialization method for 'PyLocalObjectMeta'."""
            self._parent = parent
            self._name = name
            self._api_helper = api_helper(self)
            self._command_names = []
            self.type = "object"

            def update(clss):
                for name, cls in clss.__dict__.items():
                    if cls.__class__.__name__ in ("PyLocalCommandMeta"):
                        self._command_names.append(name)

                    if cls.__class__.__name__ in (
                        "PyLocalPropertyMeta",
                        "PyLocalObjectMeta",
                        "PyLocalCommandMeta",
                    ):
                        setattr(
                            self,
                            name,
                            cls(self, api_helper, name),
                        )
                    if (
                        cls.__class__.__name__ == "PyLocalNamedObjectMeta"
                        or cls.__class__.__name__ == "PyLocalNamedObjectMetaAbstract"
                    ):
                        setattr(
                            self,
                            cls.PLURAL,
                            PyLocalContainer(self, cls, api_helper, cls.PLURAL),
                        )
                    if cls.__class__.__name__ == "PyReferenceObjectMeta":
                        setattr(
                            self,
                            name,
                            cls(self, cls.PATH, cls.LOCATION, cls.SESSION, name),
                        )
                for base_class in clss.__bases__:
                    update(base_class)

            update(self.__class__)

        return wrapper

    @classmethod
    def __create_getattribute(cls):
        def wrapper(self, name):
            obj = object.__getattribute__(self, name)
            return obj

        return wrapper

    @classmethod
    def __create_updateitem(cls):
        def wrapper(self, value):
            properties = value
            sort_by = None
            if hasattr(self, "sort_by"):
                sort_by = self.sort_by
            elif hasattr(self, "include"):
                sort_by = self.include
            if sort_by:
                sorted_properties = {
                    prop: properties[prop] for prop in sort_by if prop in properties
                }
                sorted_properties.update(
                    {k: v for k, v in properties.items() if k not in sort_by}
                )
                properties.clear()
                properties.update(sorted_properties)
            for name, val in properties.items():
                obj = getattr(self, name)
                if obj.__class__.__class__.__name__ == "PyLocalPropertyMeta":
                    obj.set_state(val)
                else:
                    if obj.__class__.__class__.__name__ == "PyReferenceObjectMeta":
                        obj = obj.ref
                    obj.update(val)

        wrapper.__doc__ = "Update object."
        return wrapper

    @classmethod
    def __create_get_state(cls):
        def wrapper(self, show_attributes=False):
            state = {}

            if not getattr(self, "is_active", True):
                return

            def update_state(clss):
                for name, cls in clss.__dict__.items():
                    o = getattr(self, name)
                    if o is None or name.startswith("_") or name.startswith("__"):
                        continue

                    if cls.__class__.__name__ == "PyReferenceObjectMeta":
                        if o.LOCATION == "local":
                            o = o.ref
                        else:
                            continue
                    elif cls.__class__.__name__ == "PyLocalCommandMeta":
                        args = {}
                        for arg in o._args:
                            args[arg] = getattr(o, arg)()
                        state[name] = args
                    if (
                        cls.__class__.__name__ == "PyLocalObjectMeta"
                        or cls.__class__.__name__ == "PyReferenceObjectMeta"
                    ):
                        if getattr(o, "is_active", True):
                            state[name] = o(show_attributes)
                    elif (
                        cls.__class__.__name__ == "PyLocalNamedObjectMeta"
                        or cls.__class__.__name__ == "PyLocalNamedObjectMetaAbstract"
                    ):
                        container = getattr(self, cls.PLURAL)
                        if getattr(container, "is_active", True):
                            state[cls.PLURAL] = {}
                            for child_name in container:
                                o = container[child_name]
                                if getattr(o, "is_active", True):
                                    state[cls.PLURAL][child_name] = o()

                    elif cls.__class__.__name__ == "PyLocalPropertyMeta":
                        if getattr(o, "is_active", True):
                            state[name] = o()
                            attrs = show_attributes and getattr(o, "attributes", None)
                            if attrs:
                                for attr in attrs:
                                    state[name + "." + attr] = getattr(o, attr)

                for base_class in clss.__bases__:
                    update_state(base_class)

            update_state(self.__class__)
            return state

        return wrapper

    @classmethod
    def __create_setattr(cls):
        def wrapper(self, name, value):
            attr = getattr(self, name, None)
            if attr and attr.__class__.__class__.__name__ == "PyLocalPropertyMeta":
                attr.set_state(value)
            else:
                object.__setattr__(self, name, value)

        return wrapper

    @classmethod
    def __create_repr(cls):
        def wrapper(self):
            return pformat(self(True), depth=1, indent=2)

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["__getattribute__"] = cls.__create_getattribute()
        attrs["__init__"] = attrs.get("__init__", cls.__create_init())
        if "__call__" not in attrs:
            attrs["__call__"] = cls.__create_get_state()
        attrs["__setattr__"] = cls.__create_setattr()
        # attrs["__repr__"] = cls.__create_repr()
        attrs["update"] = cls.__create_updateitem()
        return super(PyLocalObjectMeta, cls).__new__(cls, name, bases, attrs)


class PyLocalCommandMeta(PyLocalObjectMeta):
    """Local object metaclass."""

    @classmethod
    def __create_init(cls):
        def wrapper(self, parent, api_helper, name=""):
            """Create the initialization method for 'PyLocalObjectMeta'."""
            self._parent = parent
            self._name = name
            self._api_helper = api_helper(self)
            self.type = "object"
            self._args = []
            self._command_names = []
            self._exe_cmd = getattr(self, "_exe_cmd")

            def update(clss):
                for name, cls in clss.__dict__.items():
                    if cls.__class__.__name__ in (
                        "PyLocalCommandArgMeta",
                        "PyLocalPropertyMeta",
                    ):
                        self._args.append(name)
                        setattr(
                            self,
                            name,
                            cls(self, api_helper, name),
                        )
                for base_class in clss.__bases__:
                    update(base_class)

            update(self.__class__)

        return wrapper

    @classmethod
    def __execute_command(cls):
        def wrapper(self, **kwargs):
            for arg_name, arg_value in kwargs.items():
                getattr(self, arg_name).set_state(arg_value)
            cmd_args = {}
            for arg_name in self._args:
                cmd_args[arg_name] = getattr(self, arg_name)()
            rv = self._exe_cmd(**cmd_args)
            return rv

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["__init__"] = cls.__create_init()
        attrs["__call__"] = cls.__execute_command()
        return super(PyLocalCommandMeta, cls).__new__(cls, name, bases, attrs)


class PyLocalNamedObjectMeta(PyLocalObjectMeta):
    """Metaclass for local named object classes."""

    @classmethod
    def __create_init(cls):
        def wrapper(self, name, parent, api_helper):
            """Create the initialization method for 'PyLocalNamedObjectMeta'."""
            self._name = name
            self._api_helper = api_helper(self)
            self._parent = parent
            self._command_names = []
            self.type = "object"

            def update(clss):
                for name, cls in clss.__dict__.items():
                    if cls.__class__.__name__ in ("PyLocalCommandMeta"):
                        self._command_names.append(name)

                    if cls.__class__.__name__ in (
                        "PyLocalPropertyMeta",
                        "PyLocalObjectMeta",
                        "PyLocalCommandMeta",
                    ):
                        # delete old property if overridden
                        if getattr(self, name).__class__.__name__ == name:
                            delattr(self, name)
                        setattr(
                            self,
                            name,
                            cls(self, api_helper, name),
                        )
                    elif (
                        cls.__class__.__name__ == "PyLocalNamedObjectMeta"
                        or cls.__class__.__name__ == "PyLocalNamedObjectMetaAbstract"
                    ):
                        setattr(
                            self,
                            cls.PLURAL,
                            PyLocalContainer(self, cls, api_helper, cls.PLURAL),
                        )
                    elif cls.__class__.__name__ == "PyReferenceObjectMeta":
                        setattr(
                            self, name, cls(self, cls.PATH, cls.LOCATION, cls.SESSION)
                        )
                for base_class in clss.__bases__:
                    update(base_class)

            update(self.__class__)

        return wrapper

    def __new__(cls, name, bases, attrs):
        attrs["__init__"] = cls.__create_init()
        return super(PyLocalNamedObjectMeta, cls).__new__(cls, name, bases, attrs)


class PyLocalNamedObjectMetaAbstract(ABCMeta, PyLocalNamedObjectMeta):
    """Local named object abstract metaclass."""

    pass


class PyLocalContainer(MutableMapping):
    """Local container for named objects."""

    def __init__(self, parent, object_class, api_helper, name=""):
        """Initialize the 'PyLocalContainer' object."""
        self._parent = parent
        self._name = name
        self.__object_class = object_class
        self._local_collection = {}
        self.__api_helper = api_helper
        self.type = "named-object"
        self._command_names = []

        if hasattr(object_class, "SHOW_AS_SEPARATE_OBJECT"):
            PyLocalContainer.show_as_separate_object = property(
                lambda self: self.__object_class.SHOW_AS_SEPARATE_OBJECT(self)
            )
        if hasattr(object_class, "EXCLUDE"):
            PyLocalContainer.exclude = property(
                lambda self: self.__object_class.EXCLUDE(self)
            )
        if hasattr(object_class, "INCLUDE"):
            PyLocalContainer.include = property(
                lambda self: self.__object_class.INCLUDE(self)
            )
        if hasattr(object_class, "LAYOUT"):
            PyLocalContainer.layout = property(
                lambda self: self.__object_class.LAYOUT(self)
            )
        if hasattr(object_class, "STYLE"):
            PyLocalContainer.style = property(
                lambda self: self.__object_class.STYLE(self)
            )
        if hasattr(object_class, "ICON"):
            PyLocalContainer.icon = property(
                lambda self: self.__object_class.ICON(self)
            )
        if hasattr(object_class, "IS_ACTIVE"):
            PyLocalContainer.is_active = property(
                lambda self: self.__object_class.IS_ACTIVE(self)
            )

        for name, cls in self.__class__.__dict__.items():
            if cls.__class__.__name__ in ("PyLocalCommandMeta"):
                self._command_names.append(name)

            if cls.__class__.__name__ in ("PyLocalCommandMeta"):
                setattr(
                    self,
                    name,
                    cls(self, api_helper, name),
                )

    def update(self, value):
        """Updates this object with the provided dictionary."""
        for name, val in value.items():
            o = self[name]
            o.update(val)

    def get_root(self, obj=None):
        """Returns the top-most parent object."""
        obj = self if obj is None else obj
        parent = obj
        if getattr(obj, "_parent", None):
            parent = self.get_root(obj._parent)
        return parent

    def get_session(self, obj=None):
        """Returns the session object."""
        root = self.get_root(obj)
        return root.session

    def get_path(self):
        """Path to the current object."""
        if getattr(self, "_parent", None):
            return self._parent.get_path() + "/" + self._name
        return self._name

    @property
    def path(self):
        """Path to the current object."""
        return self.get_path()

    @property
    def session(self):
        """Returns the session object."""
        return self.get_session()

    def get_session_handle(self, obj=None):
        """Returns the session-handle object."""
        root = self.get_root(obj)
        return getattr(root, "session_handle", None)

    @property
    def session_handle(self):
        """Returns the session-handle object."""
        return self.get_session_handle()

    def __iter__(self):
        return iter(self._local_collection)

    def __len__(self):
        return len(self._local_collection)

    def __getitem__(self, name):
        o = self._local_collection.get(name, None)
        if not o:
            o = self._local_collection[name] = self.__object_class(
                name, self, self.__api_helper
            )
            on_create = getattr(self._PyLocalContainer__object_class, "on_create", None)
            if on_create:
                on_create(self, name)
        return o

    def __setitem__(self, name, value):
        o = self[name]
        o.update(value)

    def __delitem__(self, name):
        del self._local_collection[name]
        on_delete = getattr(self._PyLocalContainer__object_class, "on_delete", None)
        if on_delete:
            on_delete(self, name)

    def _get_unique_chid_name(self):
        children = list(self)
        index = 0
        while True:
            unique_name = (
                f"{self._PyLocalContainer__object_class.__name__.lower()}-{index}"
            )
            if unique_name not in children:
                break
            index += 1
        return unique_name

    class Delete(metaclass=PyLocalCommandMeta):
        """Local delete command."""

        def _exe_cmd(self, names):
            for item in names:
                self._parent.__delitem__(item)

        class names(metaclass=PyLocalPropertyMeta):
            """Local names property."""

            value: List[str] = []

            @Attribute
            def allowed_values(self):
                """Get allowed values."""
                return list(self._parent._parent)

    class Create(metaclass=PyLocalCommandMeta):
        """Local create command."""

        def _exe_cmd(self, name=None):
            if name is None:
                name = self._parent._get_unique_chid_name()
            new_object = self._parent.__getitem__(name)
            return new_object._name

        class name(metaclass=PyLocalPropertyMeta):
            """Local name property."""

            value: str = None



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for file transfer service."""

import os
import pathlib
import random
import shutil
from typing import Any, Protocol
import warnings

from ansys.fluent.core.pyfluent_warnings import PyFluentUserWarning
from ansys.fluent.core.utils import get_user_data_dir
from ansys.fluent.core.utils.deprecate import deprecate_argument
import ansys.platform.instancemanagement as pypim

# Host path which is mounted to the file-transfer-service container
MOUNT_SOURCE = str(get_user_data_dir())


class PyPIMConfigurationError(ConnectionError):
    """Raised when `PyPIM<https://pypim.docs.pyansys.com/version/stable/>` is not configured."""

    def __init__(self):
        """Initialize PyPIMConfigurationError."""
        super().__init__("PyPIM is not configured.")


class FileTransferStrategy(Protocol):
    """Provides the file transfer strategy."""

    def upload(
        self, file_name: list[str] | str, remote_file_name: str | None = None
    ) -> None:
        """Upload a file to the server.

        Parameters
        ----------
        file_name : str
            File name.
        remote_file_name : str, optional
            Remote file name. The default is ``None``.
        """
        ...

    def download(
        self, file_name: list[str] | str, local_directory: str | None = None
    ) -> None:
        """Download a file from the server.

        Parameters
        ----------
        file_name : str
            File name.
        local_directory : str, optional
            Local directory. The default is ``None``.
        """
        ...


class LocalFileTransferStrategy(FileTransferStrategy):
    """Provides the local file transfer strategy can be used for Fluent launched in the
    standalone mode.

    Examples
    --------
    >>> import ansys.fluent.core as pyfluent
    >>> from ansys.fluent.core import examples
    >>> from ansys.fluent.core.utils.file_transfer_service import LocalFileTransferStrategy
    >>> mesh_file_name = examples.download_file("mixing_elbow.msh.h5", "pyfluent/mixing_elbow")
    >>> meshing_session = pyfluent.launch_fluent(mode=pyfluent.FluentMode.MESHING, file_transfer_service=LocalFileTransferStrategy())
    >>> meshing_session.upload(file_name=mesh_file_name, remote_file_name="elbow.msh.h5")
    >>> meshing_session.meshing.File.ReadMesh(FileName="elbow.msh.h5")
    >>> meshing_session.meshing.File.WriteMesh(FileName="write_elbow.msh.h5")
    >>> meshing_session.download(file_name="write_elbow.msh.h5", local_directory="<local_directory_path>")
    """

    def __init__(self, server_cwd: str | None = None):
        """Local File Transfer Service.

        Parameters
        ----------
        server_cwd: str, Optional
            Current working directory of server/Fluent.
        """
        self.pyfluent_cwd = pathlib.Path(str(os.getcwd()))
        self.fluent_cwd = (
            pathlib.Path(str(server_cwd)) if server_cwd else self.pyfluent_cwd
        )

    def file_exists_on_remote(self, file_name: str) -> bool:
        """Check if remote file exists.

        Parameters
        ----------
        file_name: str
            File name.

        Returns
        -------
            Whether file exists.
        """
        full_file_name = pathlib.Path(self.fluent_cwd) / os.path.basename(file_name)
        return full_file_name.is_file()

    def upload(
        self, file_name: list[str] | str, remote_file_name: str | None = None
    ) -> None:
        """Upload a file to the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        remote_file_name : str, optional
            Remote file name. The default is ``None``.

        Raises
        ------
        FileNotFoundError
            If a file does not exist.

        Examples
        --------
        >>> import ansys.fluent.core as pyfluent
        >>> from ansys.fluent.core import examples
        >>> from ansys.fluent.core.utils.file_transfer_service import LocalFileTransferStrategy
        >>> mesh_file_name = examples.download_file("mixing_elbow.msh.h5", "pyfluent/mixing_elbow")
        >>> meshing_session = pyfluent.launch_fluent(mode=pyfluent.FluentMode.MESHING, file_transfer_service=LocalFileTransferStrategy())
        >>> meshing_session.upload(file_name=mesh_file_name, remote_file_name="elbow.msh.h5")
        >>> meshing_session.meshing.File.ReadMesh(FileName="elbow.msh.h5")
        """
        files = _get_files(file_name)
        for file in files:
            if file.is_file():
                if remote_file_name:
                    shutil.copyfile(
                        file,
                        str(self.fluent_cwd / f"{os.path.basename(remote_file_name)}"),
                    )
                else:
                    shutil.copyfile(
                        file, str(self.fluent_cwd / f"{os.path.basename(file)}")
                    )

    def download(
        self, file_name: list[str] | str, local_directory: str | None = None
    ) -> None:
        """Download a file from the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        local_directory : str, optional
            Local directory. The default is ``None``.

        Examples
        --------
        >>> import ansys.fluent.core as pyfluent
        >>> from ansys.fluent.core import examples
        >>> from ansys.fluent.core.utils.file_transfer_service import LocalFileTransferStrategy
        >>> mesh_file_name = examples.download_file("mixing_elbow.msh.h5", "pyfluent/mixing_elbow")
        >>> meshing_session = pyfluent.launch_fluent(mode=pyfluent.FluentMode.MESHING, file_transfer_service=LocalFileTransferStrategy())
        >>> meshing_session.meshing.File.WriteMesh(FileName="write_elbow.msh.h5")
        >>> meshing_session.download(file_name="write_elbow.msh.h5", local_directory="<local_directory_path>")
        """
        files = _get_files(file_name)
        for file in files:
            remote_file_name = str(self.fluent_cwd / file.name)
            local_file_name = None
            if local_directory:
                local_dir_path = pathlib.Path(local_directory)
                if local_dir_path.is_dir():
                    local_file_name = local_dir_path / file.name
                else:
                    local_file_name = local_dir_path
            else:
                local_file_name = self.pyfluent_cwd / file.name
            if local_file_name.exists() and local_file_name.samefile(remote_file_name):
                return
            shutil.copyfile(remote_file_name, str(local_file_name))


def _get_files(
    file_name: str | pathlib.PurePath | list[str | pathlib.PurePath],
):
    if isinstance(file_name, (str, pathlib.PurePath)):
        files = [pathlib.Path(file_name)]
    elif isinstance(file_name, list):
        files = [pathlib.Path(file) for file in file_name]
    return files


class RemoteFileTransferStrategy(FileTransferStrategy):
    """Provides a file transfer service based on the `gRPC client <https://filetransfer.tools.docs.pyansys.com/version/stable/>`_
    and `gRPC server <https://filetransfer-server.tools.docs.pyansys.com/version/stable/>`_.

    Examples
    --------
    >>> import ansys.fluent.core as pyfluent
    >>> from ansys.fluent.core import examples
    >>> from ansys.fluent.core.utils.file_transfer_service import RemoteFileTransferStrategy
    >>> case_file_name = examples.download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    >>> solver_session = pyfluent.launch_fluent(file_transfer_service=RemoteFileTransferStrategy())
    >>> solver_session.upload(file_name=case_file_name, remote_file_name="elbow.cas.h5")
    >>> solver_session.file.read_case(file_name="elbow.cas.h5")
    >>> solver_session.file.write_case(file_name="write_elbow.cas.h5")
    >>> solver_session.download(file_name="write_elbow.cas.h5", local_directory="<local_directory_path>")
    """

    @deprecate_argument("container_mount_path", "mount_target")
    @deprecate_argument("host_mount_path", "mount_source")
    def __init__(
        self,
        image_name: str | None = None,
        image_tag: str | None = None,
        port: int | None = None,
        mount_target: str | None = None,
        mount_source: str | None = None,
    ):
        """Provides the gRPC-based remote file transfer strategy.

        Parameters
        ----------
        image_name: str
            Name of the image.
        image_tag : str, optional
            Tag of the image.
        port: int, optional
            Port for the file transfer service to use.
        mount_target: str | Path, optional
            Path inside the container where ``mount_source`` will be mounted to.
        mount_source: str | Path, optional
            Existing path in the host operating system that will be mounted to ``mount_target``.
        """
        import docker

        self.docker_client = docker.from_env()
        self.image_name = (
            image_name if image_name else "ghcr.io/ansys/tools-filetransfer"
        )
        self.image_tag = image_tag if image_tag else "latest"
        self.mount_target = mount_target if mount_target else "/home/container/workdir/"
        self.mount_source = mount_source if mount_source else MOUNT_SOURCE
        try:
            self.host_port = port if port else random.randint(5000, 6000)
            self.ports = {"50000/tcp": self.host_port}
            self.container = self.docker_client.containers.run(
                image=f"{self.image_name}:{self.image_tag}",
                ports=self.ports,
                detach=True,
                volumes=[f"{self.mount_source}:{self.mount_target}"],
            )
        except docker.errors.DockerException:
            self.host_port = port if port else random.randint(6000, 7000)
            self.ports = {"50000/tcp": self.host_port}
            self.container = self.docker_client.containers.run(
                image=f"{self.image_name}:{self.image_tag}",
                ports=self.ports,
                detach=True,
                volumes=[f"{self.mount_source}:{self.mount_target}"],
            )
        import ansys.tools.filetransfer as ft

        self.client = ft.Client.from_server_address(f"localhost:{self.host_port}")

    def file_exists_on_remote(self, file_name: str) -> bool:
        """Check if remote file exists.

        Parameters
        ----------
        file_name: str
            File name.

        Returns
        -------
            Whether file exists.
        """
        full_file_name = pathlib.Path(self.mount_source) / os.path.basename(file_name)
        return full_file_name.is_file()

    def upload(self, file_name: list[str] | str, remote_file_name: str | None = None):
        """Upload a file to the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        remote_file_name : str, optional
            Remote file name. The default is ``None``.

        Raises
        ------
        FileNotFoundError
            If a file does not exist.

        Examples
        --------
        >>> import ansys.fluent.core as pyfluent
        >>> from ansys.fluent.core import examples
        >>> from ansys.fluent.core.utils.file_transfer_service import RemoteFileTransferStrategy
        >>> case_file_name = examples.download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
        >>> solver_session = pyfluent.launch_fluent(file_transfer_service=RemoteFileTransferStrategy())
        >>> solver_session.upload(file_name=case_file_name, remote_file_name="elbow.cas.h5")
        >>> solver_session.file.read_case(file_name="elbow.cas.h5")
        """
        files = _get_files(file_name)
        if self.client:
            for file in files:
                is_file_on_remote = self.file_exists_on_remote(os.path.basename(file))
                if is_file_on_remote:
                    warnings.warn(
                        f"\n{file} with the same name exists at the remote location.\n",
                        PyFluentUserWarning,
                    )
                elif os.path.isfile(file) and not is_file_on_remote:
                    self.client.upload_file(
                        local_filename=file,
                        remote_filename=(
                            remote_file_name
                            if remote_file_name
                            else os.path.basename(file)
                        ),
                    )
                else:
                    raise FileNotFoundError(f"{file} does not exist.")

    def download(self, file_name: list[str] | str, local_directory: str | None = None):
        """Download a file from the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        local_directory : str, optional
            Local directory. The default is ``None``.

        Examples
        --------
        >>> import ansys.fluent.core as pyfluent
        >>> from ansys.fluent.core import examples
        >>> from ansys.fluent.core.utils.file_transfer_service import RemoteFileTransferStrategy
        >>> case_file_name = examples.download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
        >>> solver_session = pyfluent.launch_fluent(file_transfer_service=RemoteFileTransferStrategy())
        >>> solver_session.file.write_case(file_name="write_elbow.cas.h5")
        >>> solver_session.download(file_name="write_elbow.cas.h5", local_directory="<local_directory_path>")
        """
        files = _get_files(file_name)
        if self.client:
            for file in files:
                if os.path.isfile(file):
                    warnings.warn(
                        f"\nFile already exists. File path:\n{file}\n",
                        PyFluentUserWarning,
                    )
                else:
                    self.client.download_file(
                        remote_filename=os.path.basename(file),
                        local_filename=(
                            local_directory
                            if local_directory
                            else os.path.basename(file)
                        ),
                    )


class PimFileTransferService:
    """Provides a file transfer service based on `PyPIM <https://pypim.docs.pyansys.com/version/stable/>`_ and the ``simple_upload_server()`` method.

    Attributes
    ----------
    pim_instance: PIM instance
        Instance of PIM which supports upload server services.

    file_service: Client instance
        Instance of Client which supports upload and download methods.

    Methods
    -------
    upload(
        file_name, remote_file_name
        )
        Upload a file to the server.

    download(
        file_name, local_directory
        )
        Download a file from the server.
    """

    def __init__(self, pim_instance: Any | None = None):
        """Initialize PimFileTransferService.

        Parameters
        ----------
        pim_instance: Any, optional
            PIM instance.
        """
        self.pim_instance = pim_instance
        self.upload_server = None
        self.file_service = None
        try:
            if "http-simple-upload-server" in self.pim_instance.services:
                self.upload_server = self.pim_instance.services[
                    "http-simple-upload-server"
                ]
            elif "grpc" in self.pim_instance.services:
                self.upload_server = self.pim_instance.services["grpc"]
        except (AttributeError, KeyError):
            pass
        else:
            try:
                from simple_upload_server.client import Client

                self.file_service = Client(
                    token="token",
                    url=self.upload_server.uri,
                    headers=self.upload_server.headers,
                )
            except ModuleNotFoundError:
                pass

    def file_exists_on_remote(self, file_name: str) -> bool:
        """Check if remote file exists.

        Parameters
        ----------
        file_name: str
            File name.

        Returns
        -------
            Whether file exists.
        """
        return self.file_service.file_exist(os.path.basename(file_name))

    def is_configured(self):
        """Check pypim configuration."""
        return pypim.is_configured()

    def upload_file(self, file_name: str, remote_file_name: str | None = None):
        """Upload a file to the server supported by `PyPIM<https://pypim.docs.pyansys.com/version/stable/>`.

        Parameters
        ----------
        file_name : str
            File name.
        remote_file_name : str, optional
            Remote file name. The default is ``None``.

        Raises
        ------
        FileNotFoundError
            If the file does not exist.
        PyPIMConfigurationError
            If PyPIM is not configured.
        """
        if not self.is_configured():
            raise PyPIMConfigurationError()
        elif self.file_service:
            if os.path.isfile(file_name):
                expanded_file_path = os.path.expandvars(file_name)
                upload_file_name = remote_file_name or os.path.basename(
                    expanded_file_path
                )
                self.file_service.upload_file(expanded_file_path, upload_file_name)
            else:
                raise FileNotFoundError(f"{file_name} does not exist.")

    def upload(self, file_name: list[str] | str, remote_file_name: str | None = None):
        """Upload a file to the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        remote_file_name : str, optional
            Remote file name. The default is ``None``.

        Raises
        ------
        FileNotFoundError
            If a file does not exist.
        """
        files = [file_name] if isinstance(file_name, str) else file_name
        if self.is_configured():
            for file in files:
                if os.path.isfile(file):
                    if not self.file_service.file_exist(os.path.basename(file)):
                        self.upload_file(
                            file_name=file, remote_file_name=remote_file_name
                        )
                        print(f"\n{os.path.basename(file_name)} uploaded.\n")
                    else:
                        warnings.warn(
                            f"\n{file} with the same name exists at the remote location.\n",
                            PyFluentUserWarning,
                        )
                elif not self.file_service.file_exist(os.path.basename(file)):
                    raise FileNotFoundError(f"{file} does not exist.")

    def download_file(self, file_name: str, local_directory: str | None = None):
        """Download a file from the server supported by `PyPIM<https://pypim.docs.pyansys.com/version/stable/>`.

        Parameters
        ----------
        file_name : str
            File name.
        local_directory : str, optional
            local directory, by default None

        Raises
        ------
        FileNotFoundError
            If the remote file does not exist.
        PyPIMConfigurationError
            If PyPIM is not configured.
        """
        if not self.is_configured():
            raise PyPIMConfigurationError()
        elif self.file_service:
            if self.file_service.file_exist(file_name):
                self.file_service.download_file(file_name, local_directory)
            else:
                raise FileNotFoundError("Remote file does not exist.")

    def download(self, file_name: list[str] | str, local_directory: str | None = "."):
        """Download a file from the server.

        Parameters
        ----------
        file_name : list[str] | str
            File name.
        local_directory : str, optional
            Local directory. The default is the current working directory.
        """
        files = [file_name] if isinstance(file_name, str) else file_name
        if self.is_configured():
            for file in files:
                if os.path.isfile(file):
                    warnings.warn(
                        f"\nFile already exists. File path:\n{file}\n",
                        PyFluentUserWarning,
                    )
                else:
                    self.download_file(
                        file_name=os.path.basename(file),
                        local_directory=local_directory,
                    )
                    print(f"\n{os.path.basename(file_name)} downloaded.\n")

    def __call__(self, pim_instance: Any | None = None):
        self.pim_instance = pim_instance



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A module that provides machine list construction for distributed parallel
environments, including queueing systems.

Currently supports UGE, LSF, PBS and SLURM by parsing the contents of the PE_HOSTFILE,
LSB_MCPU_HOSTS, PBS_NODEFILE and SLURM_JOB_NODELIST variables, respectively.
"""

import csv
import os
from pathlib import Path
import subprocess
from typing import Dict, List

from ansys.fluent.core.scheduler.machine_list import Machine, MachineList


def load_machines(
    machine_info: List[Dict[str, int]] | None = None,
    host_info: str | None = None,
    ncores: int | None = None,
) -> MachineList:
    """Provide a function to construct a machine list from allocated machines.

    Parameters
    ----------
    machine_info : list[dict[str, int]], optional
        List of machines provided by the caller.  Must be of the form:
            [{'machine-name' : <m-name-1>, 'core-count' : <int>},
             {'machine-name' : <m-name-2>, 'core-count' : <int>},
             ... ]
    host_info : str, optional
        Host file name or list of machines and cores as a string separated by
        commas and colons as follows:
        Example 1:  'M0:3,M1:2'
        Example 2:  'M0,M0,M0,M1,M1'
    ncores : int, optional
        Total core count.
        If provided without `machine_info`, sets the core count for local
        parallel. If both `machine_info` and `ncores` are provided, then the
        machine list determined by `machine_info` will be limited by the
        `ncores` value.

    Returns
    -------
    MachineList
        A list of machines.

    Notes
    -----
    On UGE the PE_HOSTFILE variable is used to find machines, LSB_MCPU_HOSTS
    list for LSF, PBS_NODEFILE for PBS and SLURM_JOB_NODELIST on SLURM.
    Unsupported job schedulers may provide alternative ways of providing a list
    of machines, in that case the list must be pre-parsed and provided via the
    `machine_info` or `host_info` parameters.

    In some SLURM environments, the hostnames contained within the variable
    SLURM_JOB_NODELIST may not be valid to ssh to. In that case those names
    cannot be passed to the solver. So, in the SLURM branch there is a test to
    check if ssh to the first host is working, and if not, get 'actual' machine
    names using scontrol.
    """

    machine_list = MachineList()

    if machine_info:
        machine_list = _construct_machine_list_manual(machine_info)
    elif host_info:
        machine_list = _parse_host_info(host_info)
    elif "PE_HOSTFILE" in os.environ:
        hostfile_name = os.environ.get("PE_HOSTFILE")
        machine_list = _construct_machine_list_uge(hostfile_name)
    elif "LSB_MCPU_HOSTS" in os.environ:
        hostList = os.environ.get("LSB_MCPU_HOSTS")
        machine_list = _construct_machine_list_lsf(hostList)
    elif "PBS_NODEFILE" in os.environ:
        hostfile_name = os.environ.get("PBS_NODEFILE")
        machine_list = _construct_machine_list_pbs(hostfile_name)
    elif "SLURM_JOB_NODELIST" in os.environ:
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machine_list = _construct_machine_list_slurm(hostList)
        sshTest = (
            "ssh "
            + str(machine_list.machines[0].host_name)
            + " /bin/true > /dev/null 2>&1; echo $?"
        )
        p = subprocess.Popen(sshTest, shell=True, stdout=subprocess.PIPE)
        procOutput = p.communicate()
        if procOutput[0] != b"0\n":
            runCommand = (
                r"scontrol show node ${SLURM_JOB_NODELIST} | "
                r"awk '/NodeAddr=/ {print $1}' | cut -f2 -d="
            )
            p = subprocess.Popen(runCommand, shell=True, stdout=subprocess.PIPE)
            procOutput = p.communicate()
            hostList = procOutput[0].decode("utf-8").replace("\n", ",")
            length = len(hostList)
            hostList = hostList[0 : length - 1]
            machine_list = _construct_machine_list_slurm(hostList)
    elif "CCP_NODES" in os.environ:
        hostList = os.environ.get("CCP_NODES")
        machine_list = _construct_machine_list_ccs(hostList)
    elif ncores:
        machine_list = _get_local_machine(ncores)
    elif ncores is None:
        machine_list = _get_local_machine(1)

    if ncores is not None and ncores < machine_list.number_of_cores:
        # If both machine list and number of cores are provided, edit the
        # machine list to use exactly the number of cores indicated.
        machine_list = _restrict_machines_to_core_count(machine_list, ncores)

    return machine_list


def _parse_host_info(host_info):
    """Parse the host machine information from command-line arguments.

    Returns
    -------
    list[dict] :
        A list of dictionaries formatted as:
        {'machine-name' : ###, 'core-count' : ###}
    """
    if Path(host_info).is_file():
        # Only opens a file if it exists
        with open(host_info, "r") as f:
            host_info = f.read()

    sMod = 1 if host_info[0] == "[" else 0
    sBeg = sMod
    sEnd = len(host_info) - sMod
    machine_data = host_info[sBeg:sEnd].split(",")

    return _parse_machine_data(machine_data)


def _parse_machine_data(machine_data):
    """Parse the host machine data provided as a list of strings.

    Parameters
    ----------
    machine_data : list[str]
        The data can be a list of machines such as:
            ["M0","M1","M1"]
        or it can include cores as well:
            ["M0:1","M1:2"]
        or a combination thereof.

    Returns
    -------
    list[dict] :
        The return value is a list of dictionaries formatted as:
        {'machine-name' : ###, 'core-count' : ###}
    """
    machineList = MachineList()

    for datum in machine_data:
        # Parse machine name and core count
        if ":" in datum:
            # Machine and core given
            datumPair = datum.split(":")
            machineName = datumPair[0].strip()
            ncores = int(datumPair[1])
        else:
            # Just machine name - assume one core
            machineName = datum.strip()
            ncores = 1

        if machineName == "":
            raise RuntimeError("Problem with machine list format.")

        # Add to existing machine if already in the list
        for machine in machineList.machines:
            if machine.host_name == machineName:
                machine.number_of_cores += ncores
                break
        else:  # machine name not already in machineList
            machineList.add(Machine(machineName, ncores))

    return machineList


def _get_local_machine(ncores):
    """Provide private module function to convert a core count into a machine list for a
    local job."""

    import socket

    hostName = socket.gethostname()
    machineList = MachineList()
    machineList.add(Machine(hostName, ncores))

    return machineList


def _restrict_machines_to_core_count(old_machine_list, ncores):
    """Provide private module function to adjust the number of cores used per machine
    based on a user-supplied core count.

    Parameters
    ----------
    old_machine_list : MachineList
        List of machines to be modified.
    ncores : int
        Updated core count.

    Returns
    -------
    MachineList
       New MachineList constrained to number of requested cores.

    Notes
    -----
    Every machine contributes cores to the new list until ncores is reached.
    However, the original machine order is preserved.  This ensures that all
    machines are maximally utilized.  The old machine list is sorted by core
    count so that uneven distributions favor machines with more cores. If a
    total of x cores are available on the machines and x <= ncores, then the
    returned machine list will be identical to the input.
    """
    if ncores >= old_machine_list.number_of_cores:
        return old_machine_list

    # Get indices ordering the machines from largest to smallest core count
    machineListOrder = [
        i[0]
        for i in sorted(
            enumerate(old_machine_list.machines),
            key=lambda x: x[1].number_of_cores,
            reverse=True,
        )
    ]

    newMachineList = MachineList()
    for m in old_machine_list.machines:
        newMachineList.add(Machine(m.host_name, 0, m.queue_name, m.core_list))

    ncoresRemain = ncores
    while ncoresRemain != 0:
        for i in machineListOrder:
            oldMachine = old_machine_list.machines[i]
            if oldMachine.number_of_cores != 0 and ncoresRemain != 0:
                newMachineList.machines[i].number_of_cores += 1
                oldMachine.number_of_cores -= 1
                ncoresRemain -= 1
                if ncoresRemain == 0:
                    break

    return newMachineList


def _construct_machine_list_uge(host_file_name):
    """Provide private module function to parse the UGE host file."""
    csv.register_dialect("pemachines", delimiter=" ", skipinitialspace=True)
    machineList = MachineList()
    with open(host_file_name, "r") as peFile:
        peReader = csv.reader(peFile, dialect="pemachines")
        for row in peReader:
            if len(row) == 0:
                continue
            if len(row) == 1:
                row.append(1)
            else:
                row[1] = int(row[1])
            machineList.add(Machine(*row))
    return machineList


def _construct_machine_list_lsf(host_list):
    """Provide private module function to parse the LSF host list."""
    machineList = MachineList()
    splitHostList = host_list.split()
    im = 0
    while im < len(splitHostList):
        machineList.add(Machine(splitHostList[im], int(splitHostList[im + 1])))
        im += 2
    return machineList


def _construct_machine_list_pbs(host_file_name):
    """Provide private module function to parse the PBS host file."""
    # PBS_NODE file has one machine name per line per core allocated on the machine.
    # It's identical to a Fluent host file format.  This code accumulates the total
    # core count on each machine.
    machineDict = {}
    with open(host_file_name, "r") as pbsFile:
        for hostname in pbsFile:
            hostname = hostname.rstrip("\r\n")
            if len(hostname) == 0:
                continue
            if hostname in machineDict:
                machineDict[hostname] += 1
            else:
                machineDict[hostname] = 1

    # Convert accumulated dictionary to a MachineList
    machineList = MachineList()
    for m in list(machineDict):
        machine = Machine(m, machineDict[m])
        machineList.add(machine)
    return machineList


def _construct_machine_list_slurm(host_list):
    """Provide a private module function to parse the SLURM host and task lists.

    The SLURM system provides a comma separated list of host names.  The host
    names may be listed individually or consecutive host names may have IDs that
    are provided by a set within brackets:

    SLURM_JOB_NODELIST = machinea[2-5,7,14-15],machineb,machinec[008-010,012,017-019],machined[099-101] ...

    Consecutive IDs may be prefixed (or pre-padded) with zeros so that the
    string representation of each machine ID always has the same length as the
    number of digits required to represent the last machine ID in the bracketed
    range.

    The cores allocated to each machine come in a separate variable

    SLURM_TASKS_PER_NODE = '10,3,12(x2),4,15(x5)'

    An (x#) after the core count indicates that the core count is repeated #
    times.  The order is the same as SLURM_JOB_NODELIST.
    """
    import re

    machineList = MachineList()
    splitHostList = host_list.split(",")
    coresPerMachine = 1
    ntasksPerNodeSet = False
    if "SLURM_NTASKS_PER_NODE" in os.environ:
        coresPerMachine = int(os.environ.get("SLURM_NTASKS_PER_NODE"))
        ntasksPerNodeSet = True

    # Regular expression to identify if a host entry contains a single range of machines
    pRange = re.compile(r"\[.*\]")
    # Regular expressions to identify a single machine ID within brackets
    pIDOne = re.compile(r"^.*\[(\d*).$")
    pIDOneNext = re.compile(r"^(\d*)")
    # Regular expressions to identify a range of machine IDs within brackets
    pIDRangeFirst = re.compile(r"^.*\[(\d*)-(\d*).*$")
    pIDRangeNext = re.compile(r"^(\d*)-(\d*)")
    # Regular expressions to identify if the IDs in a range use zero padding
    pIdsPadded = re.compile(r"^.*\[(0\d*)-\d*.*$")
    pIdsPaddedNext = re.compile(r"(0\d*)-\d*")
    # Regular expression to identify the machine name prefix for a range
    pMachinePrefix = re.compile(r"(^.*)\[")

    entry = 0
    while entry < len(splitHostList):
        hosts = splitHostList[entry]
        prefixMatch = pMachinePrefix.match(hosts)
        # Machine has no brackets, just add to the list
        if not prefixMatch:
            machineList.add(Machine(hosts, coresPerMachine))
            entry += 1
        # Add all machines in the bracketed range if one is provided
        else:
            machinePrefix = prefixMatch.group(1)
            # Check if first bracket entry is "M[a-b" or "M[a".  Check for a range first.
            machineIDs = pIDRangeFirst.match(hosts)
            if machineIDs:
                idfirst = int(machineIDs.group(1))
                idlast = int(machineIDs.group(2))
                paddedIDs = pIdsPadded.match(hosts)
                for id in range(idfirst, idlast + 1):
                    if paddedIDs:
                        machineName = machinePrefix + str(id).rjust(
                            len(paddedIDs.group(1)), "0"
                        )
                    else:
                        machineName = machinePrefix + str(id)
                    machineList.add(Machine(machineName, coresPerMachine))
            else:
                machineIDs = pIDOne.match(hosts)
                id = int(machineIDs.group(1))
                numch = len(re.compile(r"^.*\[(\d*).$").match(hosts).group(1))
                machineName = machinePrefix + str(id).rjust(numch, "0")
                machineList.add(Machine(machineName, coresPerMachine))

            entry += 1
            # If a host has more than one numbered range, process them.
            if len(pRange.findall(hosts)) == 0:
                if entry < len(splitHostList):
                    hosts = splitHostList[entry]
                    # Check if next entry is "a-b" or "a".  Check for a range first.
                    machineIDs = pIDRangeNext.match(hosts)
                    if machineIDs:
                        singleID = False
                    else:
                        singleID = True
                        machineIDs = pIDOneNext.match(hosts)
                    while machineIDs:
                        if singleID:
                            id = int(machineIDs.group(0))
                            numch = len(re.compile(r"^(\d*)").match(hosts).group(0))
                            machineName = machinePrefix + str(id).rjust(numch, "0")
                            machineList.add(Machine(machineName, coresPerMachine))
                        else:
                            idfirst = int(machineIDs.group(1))
                            idlast = int(machineIDs.group(2))
                            paddedIDs = pIdsPaddedNext.match(hosts)
                            for id in range(idfirst, idlast + 1):
                                if paddedIDs:
                                    machineName = machinePrefix + str(id).rjust(
                                        len(paddedIDs.group(1)), "0"
                                    )
                                else:
                                    machineName = machinePrefix + str(id)
                                machineList.add(Machine(machineName, coresPerMachine))

                        entry += 1
                        if entry < len(splitHostList):
                            hosts = splitHostList[entry]
                            machineIDs = pIDRangeNext.match(hosts)
                            if machineIDs:
                                singleID = False
                            else:
                                machineIDs = pIDOneNext.match(hosts)
                                if machineIDs and len(machineIDs.group(0)) > 0:
                                    singleID = True
                                else:
                                    singleID = False
                                    machineIDs = None
                        else:
                            machineIDs = None

    if not ntasksPerNodeSet and "SLURM_TASKS_PER_NODE" in os.environ:
        splitCoreList = os.environ["SLURM_TASKS_PER_NODE"].split(",")
        coresPerMachine = []
        for numcores in splitCoreList:
            beg = numcores.find("(x")
            if beg > 0:
                end = numcores.find(")")
                for _ in range(int(numcores[beg + 2 : end])):
                    coresPerMachine.append(int(numcores[0:beg]))
            else:
                coresPerMachine.append(int(numcores))
        icores = 0
        for machine in machineList.machines:
            machine.number_of_cores = coresPerMachine[icores]
            icores += 1

    return machineList


def _construct_machine_list_ccs(host_list):
    """Provide private module function to parse the Windows HPC/CCS host list.

    Parameters
    ----------
    host_list : str
        A single string with the following format:

        "#hosts host1 #cores1 host2 #cores2 host3 #cores3 ... hostN #coresN"
    """
    machineList = MachineList()
    splitHostList = host_list.split()
    numMachines = int(splitHostList[0])
    im = 1
    for _ in range(numMachines):
        machineList.add(Machine(splitHostList[im], int(splitHostList[im + 1])))
        im += 2
    return machineList


def _construct_machine_list_manual(machine_info):
    """Provide a private module function to convert a machine information list into a
    list of machine objects."""
    machineList = MachineList()
    for m in machine_info:
        machineList.add(Machine(m["machine-name"], m["core-count"]))
    return machineList



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing tool for walking (generated) API class hierarchy.

Example
-------

.. code-block:: python

    >>> from ansys.fluent.core.codegen import walk_api
    >>> from ansys.fluent.core.generated.solver import settings_252
    >>> walk_api.walk_api(settings_252.root, lambda p: print(p), current_path=[])
    >>> walk_api.walk_api(settings_252.root, lambda p, api_item_type: print(p, api_item_type), current_path=[])

"""

from inspect import signature
from typing import List

import ansys.fluent.core.solver.flobject as flobject


def walk_api(
    api_cls, on_each_path, current_path: str | List[str] = "", api_item_type: str = ""
):
    """
    Recursively traverse the API hierarchy, calling `on_each_path` for each item.

    Parameters:
    - api_cls: The current class of the API hierarchy.
    - on_each_path: A callback function to call for each path.
    - current_path: The current path in the hierarchy (default: empty string).
      Paths can be either dot-separated strings or string lists. The type is
      determined by the client.
    """
    # Skip the root path
    if current_path:
        if len(signature(on_each_path).parameters) == 3:
            on_each_path(current_path, api_item_type, api_cls)
        elif len(signature(on_each_path).parameters) == 2:
            on_each_path(current_path, api_item_type)
        else:
            on_each_path(current_path)

    child_classes = getattr(api_cls, "_child_classes", {})

    def _traverse(child_name, api_item_type):
        if child_name in child_classes:
            child_cls = child_classes[child_name]
            # Construct the new path
            if isinstance(current_path, list):
                new_path = current_path + [child_name]
            else:
                new_path = (
                    f"{current_path}.{child_name}" if current_path else child_name
                )
            # Recursively walk the child
            if not api_item_type:
                api_item_type = (
                    "parameter"
                    if isinstance(child_cls, flobject.Property)
                    else "object"
                )
            walk_api(child_cls, on_each_path, new_path, api_item_type)

            # Delegate directly to any child_object_type (relevant for named objects)
            child_object_type = getattr(api_cls, "child_object_type", None)
            if child_object_type:
                walk_api(child_cls, on_each_path, current_path, api_item_type)

    # Get child names and their respective classes
    if api_item_type == "method":
        arg_names = [
            name for attr in ("argument_names",) for name in getattr(api_cls, attr, [])
        ]

        for arg_name in arg_names:
            _traverse(arg_name, api_item_type="argument")

    else:

        child_names = [
            name for attr in ("child_names",) for name in getattr(api_cls, attr, [])
        ]

        for child_name in child_names:
            _traverse(child_name, api_item_type="")

        method_names = [
            name
            for attr in ("command_names", "query_names")
            for name in getattr(api_cls, attr, [])
        ]

        for method_name in method_names:
            _traverse(method_name, api_item_type="method")



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_solver_models(mixing_elbow_settings_session):
    solver_session = mixing_elbow_settings_session
    models = solver_session.setup.models
    assert models.energy.enabled()
    models.energy.enabled = False
    assert not models.energy.enabled()
    models.multiphase.models = "vof"
    assert models.multiphase.models() == "vof"
    models.viscous.model = "laminar"
    assert models.viscous.model() == "laminar"
    models.viscous.model = "k-epsilon"
    assert models.viscous.model() == "k-epsilon"

    models.viscous.near_wall_treatment.wall_treatment = "enhanced-wall-treatment"
    models.multiphase.models = "eulerian"
    assert models.multiphase.models() == "eulerian"


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_disk_2d_models(disk_settings_session):
    solver_session = disk_settings_session
    models = solver_session.setup.models
    solver_session.setup.general.solver.two_dim_space = "axisymmetric"
    solver_session.setup.general.solver.two_dim_space = "swirl"

    models.viscous.model = "k-epsilon"
    near_wall = models.viscous.near_wall_treatment
    assert models.viscous.model() == "k-epsilon"
    near_wall.wall_treatment = "enhanced-wall-treatment"
    assert near_wall.wall_treatment() == "enhanced-wall-treatment"
    assert near_wall.wall_treatment.get_attr("allowed-values") == [
        "standard-wall-fn",
        "non-equilibrium-wall-fn",
        "enhanced-wall-treatment",
        "menter-lechner",
        "scalable-wall-functions",
    ]

    near_wall.wall_treatment = "standard-wall-fn"
    assert near_wall.wall_treatment() == "standard-wall-fn"
    near_wall.wall_treatment = "non-equilibrium-wall-fn"
    assert near_wall.wall_treatment() == "non-equilibrium-wall-fn"
    near_wall.wall_treatment = "menter-lechner"
    assert near_wall.wall_treatment() == "menter-lechner"
    near_wall.wall_treatment = "scalable-wall-functions"
    assert near_wall.wall_treatment() == "scalable-wall-functions"

    models.viscous.model = "k-omega"
    assert models.viscous.model() == "k-omega"
    assert models.viscous.k_omega_model.get_attr("allowed-values") == [
        "standard",
        "geko",
        "bsl",
        "sst",
    ]

    assert models.viscous.k_omega_model() == "standard"
    models.viscous.k_omega_model = "geko"
    assert models.viscous.k_omega_model() == "geko"
    models.viscous.k_omega_model = "bsl"
    assert models.viscous.k_omega_model() == "bsl"
    models.viscous.k_omega_model = "sst"
    assert models.viscous.k_omega_model() == "sst"
    assert models.viscous.model.get_attr("allowed-values") == [
        "inviscid",
        "laminar",
        "k-epsilon",
        "k-omega",
        "mixing-length",
        "spalart-allmaras",
        "k-kl-w",
        "transition-sst",
        "reynolds-stress",
        "scale-adaptive-simulation",
        "detached-eddy-simulation",
    ]
    k_omega_options = models.viscous.k_omega_options
    k_omega_options.kw_low_re_correction = True
    assert k_omega_options.kw_low_re_correction() is True

    turb_options = models.viscous.options
    turb_options.production_kato_launder_enabled = True
    assert turb_options.production_kato_launder_enabled() is True
    turb_options.production_limiter.clip_factor = 9
    assert turb_options.production_limiter.clip_factor() == 9
    turb_expert = models.viscous.turbulence_expert
    turb_expert.turb_non_newtonian = True
    assert turb_expert.turb_non_newtonian() is True

    models.viscous.model = "laminar"
    assert models.viscous.model() == "laminar"
    models.viscous.model = "spalart-allmaras"
    assert models.viscous.model() == "spalart-allmaras"



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching Fluent.

This module supports both starting Fluent locally and connecting to a remote instance
with gRPC.
"""

import inspect
import logging
import os
from typing import Any, Dict

import ansys.fluent.core as pyfluent
from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.launcher.container_launcher import DockerLauncher
from ansys.fluent.core.launcher.launcher_utils import (
    _confirm_watchdog_start,
    is_windows,
)
from ansys.fluent.core.launcher.pim_launcher import PIMLauncher
from ansys.fluent.core.launcher.pyfluent_enums import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    LaunchMode,
    Precision,
    UIMode,
    _get_fluent_launch_mode,
    _get_running_session_mode,
)
from ansys.fluent.core.launcher.server_info import _get_server_info
from ansys.fluent.core.launcher.slurm_launcher import SlurmFuture, SlurmLauncher
from ansys.fluent.core.launcher.standalone_launcher import StandaloneLauncher
import ansys.fluent.core.launcher.watchdog as watchdog
from ansys.fluent.core.pyfluent_warnings import PyFluentDeprecationWarning
from ansys.fluent.core.session_meshing import Meshing
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver
from ansys.fluent.core.session_solver_icing import SolverIcing
from ansys.fluent.core.utils.deprecate import deprecate_argument
from ansys.fluent.core.utils.fluent_version import FluentVersion

_THIS_DIR = os.path.dirname(__file__)
_OPTIONS_FILE = os.path.join(_THIS_DIR, "fluent_launcher_options.json")
logger = logging.getLogger("pyfluent.launcher")


def create_launcher(fluent_launch_mode: LaunchMode = None, **kwargs):
    """Use the factory function to create a launcher for supported launch modes.

    Parameters
    ----------
    fluent_launch_mode: LaunchMode
        Supported Fluent launch modes. Options are ``"LaunchMode.CONTAINER"``,
        ``"LaunchMode.PIM"``, ``"LaunchMode.SLURM"``, and ``"LaunchMode.STANDALONE"``.
    kwargs : Any
        Keyword arguments.
    Returns
    -------
    launcher: DockerLauncher | PimLauncher | StandaloneLauncher
        Session launcher.
    Raises
    ------
    DisallowedValuesError
        If an unknown Fluent launch mode is passed.
    """
    if fluent_launch_mode == LaunchMode.STANDALONE:
        return StandaloneLauncher(**kwargs)
    elif fluent_launch_mode == LaunchMode.CONTAINER:
        return DockerLauncher(**kwargs)
    elif fluent_launch_mode == LaunchMode.PIM:
        return PIMLauncher(**kwargs)
    elif fluent_launch_mode == LaunchMode.SLURM:
        return SlurmLauncher(**kwargs)


def _show_gui_to_ui_mode(old_arg_val, **kwds):
    start_container = kwds.get("start_container")
    container_dict = kwds.get("container_dict")
    if old_arg_val is True:
        if start_container is True:
            return UIMode.NO_GUI
        elif container_dict:
            return UIMode.NO_GUI
        elif os.getenv("PYFLUENT_LAUNCH_CONTAINER") == "1":
            return UIMode.NO_GUI
        else:
            return UIMode.GUI
    elif not old_arg_val:
        if is_windows():
            return UIMode.HIDDEN_GUI
        elif not is_windows():
            return UIMode.NO_GUI
        else:
            return None


def _version_to_dimension(old_arg_val):
    if old_arg_val == "2d":
        return Dimension.TWO
    elif old_arg_val == "3d":
        return Dimension.THREE
    else:
        return None


#   pylint: disable=unused-argument
@deprecate_argument(
    old_arg="show_gui",
    new_arg="ui_mode",
    converter=_show_gui_to_ui_mode,
    warning_cls=PyFluentDeprecationWarning,
)
@deprecate_argument(
    old_arg="version",
    new_arg="dimension",
    converter=_version_to_dimension,
    warning_cls=PyFluentDeprecationWarning,
)
def launch_fluent(
    product_version: FluentVersion | str | float | int | None = None,
    dimension: Dimension | int | None = None,
    precision: Precision | str | None = None,
    processor_count: int | None = None,
    journal_file_names: None | str | list[str] = None,
    start_timeout: int = None,
    additional_arguments: str = "",
    env: Dict[str, Any] | None = None,
    start_container: bool | None = None,
    container_dict: dict | None = None,
    dry_run: bool = False,
    cleanup_on_exit: bool = True,
    start_transcript: bool = True,
    ui_mode: UIMode | str | None = None,
    graphics_driver: (
        FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
    ) = None,
    case_file_name: str | None = None,
    case_data_file_name: str | None = None,
    lightweight_mode: bool | None = None,
    mode: FluentMode | str | None = None,
    py: bool | None = None,
    gpu: bool | list[int] | None = None,
    cwd: str | None = None,
    fluent_path: str | None = None,
    topy: str | list | None = None,
    start_watchdog: bool | None = None,
    scheduler_options: dict | None = None,
    file_transfer_service: Any | None = None,
) -> Meshing | PureMeshing | Solver | SolverIcing | SlurmFuture | dict:
    """Launch Fluent locally in server mode or connect to a running Fluent server
    instance.

    Parameters
    ----------
    product_version : FluentVersion or str or float or int, optional
        Version of Ansys Fluent to launch. To use Fluent version 2025 R1, pass
        any of  ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``or ``251``.
        The default is ``None``, in which case the newest installed version is used.
    dimension : Dimension or int, optional
        Geometric dimensionality of the Fluent simulation. The default is ``None``,
        in which case ``Dimension.THREE`` is used. Options are either the values of the
        ``Dimension`` enum (``Dimension.TWO`` or ``Dimension.THREE``) or any of ``2`` and ``3``.
    precision : Precision or str, optional
        Floating point precision. The default is ``None``, in which case ``Precision.DOUBLE``
        is used. Options are either the values of the ``Precision`` enum (``Precision.SINGLE``
        or ``Precision.DOUBLE``) or any of ``"double"`` and ``"single"``.
    processor_count : int, optional
        Number of processors. The default is ``None``, in which case ``1``
        processor is used.  In job scheduler environments the total number of
        allocated cores is clamped to value of ``processor_count``.
    journal_file_names : str or list of str, optional
        The string path to a Fluent journal file, or a list of such paths. Fluent will execute the
        journal(s). The default is ``None``.
    start_timeout : int, optional
        Maximum allowable time in seconds for connecting to the Fluent
        server. The default is ``60`` if Fluent is launched outside a Slurm environment,
        no timeout if Fluent is launched within a Slurm environment.
    additional_arguments : str, optional
        Additional arguments to send to Fluent as a string in the same
        format they are normally passed to Fluent on the command line.
    env : dict[str, str], optional
        Mapping to modify environment variables in Fluent. The default
        is ``None``.
    start_container : bool, optional
        Specifies whether to launch a Fluent Docker container image. For more details about containers, see
        :mod:`~ansys.fluent.core.launcher.fluent_container`.
    container_dict : dict, optional
        Dictionary for Fluent Docker container configuration. If specified,
        setting ``start_container = True`` as well is redundant.
        Will launch Fluent inside a Docker container using the configuration changes specified.
        See also :mod:`~ansys.fluent.core.launcher.fluent_container`.
    dry_run : bool, optional
        Defaults to False. If True, will not launch Fluent, and will instead print configuration information
        that would be used as if Fluent was being launched. If dry running a standalone start
        ``launch_fluent()`` will return a tuple containing Fluent launch string and the server info file name.
        If dry running a container start, ``launch_fluent()`` will return the configured ``container_dict``.
    cleanup_on_exit : bool, optional
        Whether to shut down the connected Fluent session when PyFluent is
        exited, or the ``exit()`` method is called on the session instance,
        or if the session instance becomes unreferenced. The default is ``True``.
    start_transcript : bool, optional
        Whether to start streaming the Fluent transcript in the client. The
        default is ``True``. You can stop and start the streaming of the
        Fluent transcript subsequently via the method calls, ``transcript.start()``
        and ``transcript.stop()`` on the session object.
    ui_mode : UIMode or str, optional
        Fluent user interface mode. Options are either the values of the ``UIMode``
        enum or any of ``"no_gui_or_graphics"``, ``"no_gui"``, ``"hidden_gui"``,
        ``"no_graphics"`` or ``"gui"``. The default is ``UIMode.HIDDEN_GUI`` in
        Windows and ``UIMode.NO_GUI`` in Linux. ``"no_gui_or_graphics"`` and
        ``"no_gui"`` user interface modes are supported in Windows starting from Fluent
        version 2024 R1.
    graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver or str, optional
        Graphics driver of Fluent. In Windows, options are either the values of the
        ``FluentWindowsGraphicsDriver`` enum or any of ``"null"``, ``"msw"``,
        ``"dx11"``, ``"opengl2"``, ``"opengl"`` or ``"auto"``. In Linux, options are
        either the values of the ``FluentLinuxGraphicsDriver`` enum or any of
       ``"null"``, ``"x11"``, ``"opengl2"``, ``"opengl"`` or ``"auto"``. The default is
       ``FluentWindowsGraphicsDriver.AUTO`` in Windows and
       ``FluentLinuxGraphicsDriver.AUTO`` in Linux.
    case_file_name : str, optional
        If provided, the case file at ``case_file_name`` is read into the Fluent session.
    case_data_file_name : str, optional
        If provided, the case and data files at ``case_data_file_name`` are read into the Fluent session.
    lightweight_mode : bool, optional
        Whether to run in lightweight mode. In lightweight mode, the lightweight settings are read into the
        current Fluent solver session. The mesh is read into a background Fluent solver session which will
        replace the current Fluent solver session once the mesh read is complete and the lightweight settings
        made by the user in the current Fluent solver session have been applied in the background Fluent
        solver session. This is all orchestrated by PyFluent and requires no special usage.
        This parameter is used only when ``case_file_name`` is provided. The default is ``False``.
    mode : str, optional
        Launch mode of Fluent to point to a specific session type.
        The default value is ``None``. Options are ``"meshing"``,
        ``"pure-meshing"`` and ``"solver"``.
    py : bool, optional
        If True, Fluent will run in Python mode. Default is None.
    gpu : bool or list, optional
        This option will start Fluent with the GPU Solver. A list of GPU IDs can be
        passed to use specific GPUs. If True is passed, the number of GPUs used will be
        clamped to the value of ``processor_count``. Please refer to
        *Starting the Fluent GPU Solver* section in *Fluent's User Guide* for more
        information like how to determine the GPU IDs.
    cwd : str, Optional
        Working directory for the Fluent client.
    fluent_path: str, Optional
        User provided Fluent installation path.
    topy : bool or str, optional
        A boolean flag to write the equivalent Python journal(s) from the journal(s) passed.
        Can optionally take the file name of the new python journal file.
    start_watchdog : bool, optional
        When ``cleanup_on_exit`` is True, ``start_watchdog`` defaults to True,
        which means an independent watchdog process is run to ensure
        that any local GUI-less Fluent sessions started by PyFluent are properly closed (or killed if frozen)
        when the current Python process ends.
    scheduler_options : dict, optional
        Dictionary containing scheduler options. Default is None.

        Currently only the Slurm scheduler is supported. The ``scheduler_options``
        dictionary must be of the form ``{"scheduler": "slurm",
        "scheduler_headnode": "<headnode>", "scheduler_queue": "<queue>",
        "scheduler_account": "<account>"}``. The keys ``scheduler_headnode``,
        ``scheduler_queue`` and ``scheduler_account`` are optional and should be
        specified in a similar manner to Fluent's scheduler options.
    file_transfer_service : optional
        File transfer service. Uploads/downloads files to/from the server.

    Returns
    -------
    :obj:`~typing.Union` [:class:`Meshing<ansys.fluent.core.session_meshing.Meshing>`, \
    :class:`~ansys.fluent.core.session_pure_meshing.PureMeshing`, \
    :class:`~ansys.fluent.core.session_solver.Solver`, \
    :class:`~ansys.fluent.core.session_solver_icing.SolverIcing`, dict]
        Session object or configuration dictionary if ``dry_run = True``.

    Raises
    ------
    UnexpectedKeywordArgument
        If an unexpected keyword argument is provided.

    Notes
    -----
    Job scheduler environments such as SLURM, LSF, PBS, etc. allocates resources / compute nodes.
    The allocated machines and core counts are queried from the scheduler environment and
    passed to Fluent.
    """
    if env is None:
        env = {}

    pyfluent.FLUENT_PRECISION_MODE = precision

    def _mode_to_launcher_type(fluent_launch_mode: LaunchMode):
        launcher_mode_type = {
            LaunchMode.CONTAINER: DockerLauncher,
            LaunchMode.PIM: PIMLauncher,
            LaunchMode.SLURM: SlurmLauncher,
            LaunchMode.STANDALONE: StandaloneLauncher,
        }
        return launcher_mode_type[fluent_launch_mode]

    argvals = inspect.getargvalues(inspect.currentframe()).locals

    fluent_launch_mode = _get_fluent_launch_mode(
        start_container=start_container,
        container_dict=container_dict,
        scheduler_options=scheduler_options,
    )

    launcher_type = _mode_to_launcher_type(fluent_launch_mode)
    launch_fluent_args = set(inspect.signature(launch_fluent).parameters.keys())
    launcher_type_args = set(
        inspect.signature(launcher_type.__init__).parameters.keys()
    )
    common_args = launch_fluent_args.intersection(launcher_type_args)
    launcher_argvals = {arg: val for arg, val in argvals.items() if arg in common_args}
    if pyfluent.START_WATCHDOG is False:
        launcher_argvals["start_watchdog"] = False
    launcher = launcher_type(**launcher_argvals)
    return launcher()


def connect_to_fluent(
    ip: str | None = None,
    port: int | None = None,
    cleanup_on_exit: bool = False,
    start_transcript: bool = True,
    server_info_file_name: str | None = None,
    password: str | None = None,
    start_watchdog: bool | None = None,
) -> Meshing | PureMeshing | Solver | SolverIcing:
    """Connect to an existing Fluent server instance.

    Parameters
    ----------
    ip : str, optional
        IP address for connecting to an existing Fluent instance. The
        IP address defaults to ``"127.0.0.1"``. You can also use the environment
        variable ``PYFLUENT_FLUENT_IP=<ip>`` to set this parameter.
        The explicit value of ``ip`` takes precedence over ``PYFLUENT_FLUENT_IP=<ip>``.
    port : int, optional
        Port to listen on for an existing Fluent instance. You can use the
        environment variable ``PYFLUENT_FLUENT_PORT=<port>`` to set a default
        value. The explicit value of ``port`` takes precedence over
        ``PYFLUENT_FLUENT_PORT=<port>``.
    cleanup_on_exit : bool, optional
        Whether to shut down the connected Fluent session when PyFluent is
        exited, or the ``exit()`` method is called on the session instance,
        or if the session instance becomes unreferenced. The default is ``False``.
    start_transcript : bool, optional
        Whether to start streaming the Fluent transcript in the client. The
        default is ``True``. You can stop and start the streaming of the
        Fluent transcript subsequently via the method calls, ``transcript.start()``
        and ``transcript.stop()`` on the session object.
    server_info_file_name: str
        Path to server-info file written out by Fluent server. The default is
        ``None``. PyFluent uses the connection information in the file to
        connect to a running Fluent session.
    password : str, optional
        Password to connect to existing Fluent instance.
    start_watchdog: bool, optional
        When ``cleanup_on_exit`` is True, ``start_watchdog`` defaults to True,
        which means an independent watchdog process is run to ensure
        that any local Fluent connections are properly closed (or terminated if frozen) when Python process ends.

    Returns
    -------
    :obj:`~typing.Union` [:class:`Meshing<ansys.fluent.core.session_meshing.Meshing>`, \
    :class:`~ansys.fluent.core.session_pure_meshing.PureMeshing`, \
    :class:`~ansys.fluent.core.session_solver.Solver`, \
    :class:`~ansys.fluent.core.session_solver_icing.SolverIcing`]
        Session object.
    """
    ip, port, password = _get_server_info(server_info_file_name, ip, port, password)
    fluent_connection = FluentConnection(
        ip=ip,
        port=port,
        password=password,
        cleanup_on_exit=cleanup_on_exit,
    )
    new_session = _get_running_session_mode(fluent_connection)

    start_watchdog = _confirm_watchdog_start(
        start_watchdog, cleanup_on_exit, fluent_connection
    )

    if start_watchdog:
        logger.info("Launching Watchdog for existing Fluent session...")
        ip, port, password = _get_server_info(server_info_file_name, ip, port, password)
        watchdog.launch(os.getpid(), port, password, ip)

    return new_session(
        fluent_connection=fluent_connection,
        scheme_eval=fluent_connection._connection_interface.scheme_eval,
        start_transcript=start_transcript,
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
import tempfile

import pytest

try:
    from ansys.fluent.core.solver import (
        Ablation,
        Battery,
        BoundaryCondition,
        BoundaryConditions,
        CalculationActivity,
        CaseModification,
        CellRegister,
        CellRegisters,
        CellZoneCondition,
        CellZoneConditions,
        Contour,
        Contours,
        Controls,
        ConvergenceConditions,
        CumulativePlots,
        CustomFieldFunctions,
        CustomVectors,
        DesignPoint,
        DesignPoints,
        DiscretePhase,
        DiscretePhaseHistogram,
        DynamicMesh,
        EChemistry,
        Energy,
        ExecuteCommands,
        ExpressionVolumes,
        FluidCellZone,
        FluidCellZones,
        FluidMaterial,
        FluidMaterials,
        Fluxes,
        General,
        Graphics,
        GroupSurfaces,
        Histogram,
        ImprintSurfaces,
        Initialization,
        Injections,
        InputParameters,
        InteriorBoundaries,
        InteriorBoundary,
        InterpolatedData,
        IsoClips,
        IsoSurfaces,
        LICs,
        LineSurfaces,
        Materials,
        Meshes,
        MeshInterfaces,
        Methods,
        Models,
        Monitor,
        Multiphase,
        NamedExpressions,
        Optics,
        OutputParameters,
        ParametricStudies,
        ParametricStudy,
        ParticleTracks,
        PartitionSurfaces,
        Pathlines,
        Pemfc,
        PlaneSlices,
        PlaneSurface,
        PlaneSurfaces,
        Plots,
        PointSurfaces,
        PressureOutlet,
        PressureOutlets,
        ProfileData,
        QuadricSurfaces,
        Radiation,
        RakeSurfaces,
        ReferenceFrame,
        ReferenceFrames,
        ReferenceValues,
        Report,
        ReportDefinitions,
        ReportFile,
        ReportFiles,
        ReportPlot,
        ReportPlots,
        Residual,
        Results,
        RunCalculation,
        SceneAnimation,
        Scenes,
        Setup,
        SimulationReports,
        Sofc,
        SolidMaterial,
        SolidMaterials,
        Solution,
        Species,
        SphereSlices,
        Structure,
        SurfaceCells,
        SurfaceIntegrals,
        Surfaces,
        SystemCoupling,
        TransformSurfaces,
        Vectors,
        VelocityInlet,
        VelocityInlets,
        VirtualBladeModel,
        Viscous,
        VolumeIntegrals,
        WallBoundaries,
        WallBoundary,
        XYPlots,
        ZoneSurfaces,
    )
except ImportError:
    pass  # for no-codegen testing workflow
import ansys.fluent.core as pyfluent
from ansys.fluent.core.examples import download_file
from ansys.fluent.core.utils.fluent_version import FluentVersion


@pytest.mark.codegen_required
def test_builtin_settings(mixing_elbow_case_data_session):
    solver = mixing_elbow_case_data_session
    fluent_version = solver.get_fluent_version()
    assert Setup(settings_source=solver) == solver.setup
    assert General(settings_source=solver) == solver.setup.general
    assert Models(settings_source=solver) == solver.setup.models
    assert Multiphase(settings_source=solver) == solver.setup.models.multiphase
    assert Energy(settings_source=solver) == solver.setup.models.energy
    assert Viscous(settings_source=solver) == solver.setup.models.viscous
    if fluent_version >= FluentVersion.v232:
        assert Radiation(settings_source=solver) == solver.setup.models.radiation
    else:
        with pytest.raises(RuntimeError):
            Radiation(settings_source=solver)
    if fluent_version >= FluentVersion.v232:
        assert Species(settings_source=solver) == solver.setup.models.species
    else:
        with pytest.raises(RuntimeError):
            Species(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert (
            DiscretePhase(settings_source=solver) == solver.setup.models.discrete_phase
        )
    else:
        with pytest.raises(RuntimeError):
            DiscretePhase(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert (
            Injections(settings_source=solver)
            == solver.setup.models.discrete_phase.injections
        )
    else:
        with pytest.raises(RuntimeError):
            Injections(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert (
            VirtualBladeModel(settings_source=solver)
            == solver.setup.models.virtual_blade_model
        )
    else:
        with pytest.raises(RuntimeError):
            VirtualBladeModel(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert Optics(settings_source=solver) == solver.setup.models.optics
    else:
        with pytest.raises(RuntimeError):
            Optics(settings_source=solver)
    if fluent_version >= FluentVersion.v232:
        assert Structure(settings_source=solver) == solver.setup.models.structure
    else:
        with pytest.raises(RuntimeError):
            Structure(settings_source=solver)
    if fluent_version >= FluentVersion.v232:
        assert Ablation(settings_source=solver) == solver.setup.models.ablation
    else:
        with pytest.raises(RuntimeError):
            Ablation(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert EChemistry(settings_source=solver) == solver.setup.models.echemistry
    else:
        with pytest.raises(RuntimeError):
            EChemistry(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert Battery(settings_source=solver) == solver.setup.models.battery
    else:
        with pytest.raises(RuntimeError):
            Battery(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert (
            SystemCoupling(settings_source=solver)
            == solver.setup.models.system_coupling
        )
    else:
        with pytest.raises(RuntimeError):
            SystemCoupling(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert Sofc(settings_source=solver) == solver.setup.models.sofc
    else:
        with pytest.raises(RuntimeError):
            Sofc(settings_source=solver)
    if fluent_version >= FluentVersion.v242:
        assert Pemfc(settings_source=solver) == solver.setup.models.pemfc
    else:
        with pytest.raises(RuntimeError):
            Pemfc(settings_source=solver)
    assert Materials(settings_source=solver) == solver.setup.materials
    assert FluidMaterials(settings_source=solver) == solver.setup.materials.fluid
    assert (
        FluidMaterial(settings_source=solver, name="air")
        == solver.setup.materials.fluid["air"]
    )
    assert SolidMaterials(settings_source=solver) == solver.setup.materials.solid
    assert (
        SolidMaterial(settings_source=solver, name="aluminum")
        == solver.setup.materials.solid["aluminum"]
    )
    assert (
        CellZoneConditions(settings_source=solver) == solver.setup.cell_zone_conditions
    )
    if fluent_version >= FluentVersion.v231:
        assert (
            CellZoneCondition(settings_source=solver, name="elbow-fluid")
            == solver.setup.cell_zone_conditions["elbow-fluid"]
        )
    else:
        with pytest.raises(RuntimeError):
            CellZoneCondition(settings_source=solver, name="elbow-fluid")
    assert (
        FluidCellZones(settings_source=solver)
        == solver.setup.cell_zone_conditions.fluid
    )
    assert (
        FluidCellZone(settings_source=solver, name="elbow-fluid")
        == solver.setup.cell_zone_conditions.fluid["elbow-fluid"]
    )
    assert (
        BoundaryConditions(settings_source=solver) == solver.setup.boundary_conditions
    )
    if fluent_version >= FluentVersion.v231:
        assert (
            BoundaryCondition(settings_source=solver, name="cold-inlet")
            == solver.setup.boundary_conditions["cold-inlet"]
        )
    else:
        with pytest.raises(RuntimeError):
            BoundaryCondition(settings_source=solver, name="cold-inlet")
    with pytest.raises(TypeError):
        BoundaryCondition(settings_source=solver, new_instance_name="bc-1")
    assert (
        VelocityInlets(settings_source=solver)
        == solver.setup.boundary_conditions.velocity_inlet
    )
    assert (
        VelocityInlet(settings_source=solver, name="cold-inlet")
        == solver.setup.boundary_conditions.velocity_inlet["cold-inlet"]
    )
    assert (
        InteriorBoundaries(settings_source=solver)
        == solver.setup.boundary_conditions.interior
    )
    assert (
        InteriorBoundary(settings_source=solver, name="interior--elbow-fluid")
        == solver.setup.boundary_conditions.interior["interior--elbow-fluid"]
    )
    assert (
        PressureOutlets(settings_source=solver)
        == solver.setup.boundary_conditions.pressure_outlet
    )
    assert (
        PressureOutlet(settings_source=solver, name="outlet")
        == solver.setup.boundary_conditions.pressure_outlet["outlet"]
    )
    assert (
        WallBoundaries(settings_source=solver) == solver.setup.boundary_conditions.wall
    )
    assert (
        WallBoundary(settings_source=solver, name="wall-elbow")
        == solver.setup.boundary_conditions.wall["wall-elbow"]
    )
    if fluent_version >= FluentVersion.v231 and fluent_version < FluentVersion.v251:
        with pytest.raises(AttributeError):
            WallBoundary(settings_source=solver, new_instance_name="wall-1")
    if fluent_version >= FluentVersion.v232:
        assert MeshInterfaces(settings_source=solver) == solver.setup.mesh_interfaces
    else:
        with pytest.raises(RuntimeError):
            MeshInterfaces(settings_source=solver)
    if fluent_version >= FluentVersion.v251:
        assert DynamicMesh(settings_source=solver) == solver.setup.dynamic_mesh
    else:
        with pytest.raises(RuntimeError):
            DynamicMesh(settings_source=solver)
    assert ReferenceValues(settings_source=solver) == solver.setup.reference_values
    if fluent_version >= FluentVersion.v232:
        assert ReferenceFrames(settings_source=solver) == solver.setup.reference_frames
    else:
        with pytest.raises(RuntimeError):
            ReferenceFrames(settings_source=solver)
    if fluent_version >= FluentVersion.v232:
        # Fluent 25.1 issue
        if fluent_version != FluentVersion.v251:
            assert (
                ReferenceFrame(settings_source=solver, name="global")
                == solver.setup.reference_frames["global"]
            )
    else:
        with pytest.raises(RuntimeError):
            ReferenceFrame(settings_source=solver, name="global")
    if fluent_version >= FluentVersion.v232:
        assert (
            NamedExpressions(settings_source=solver) == solver.setup.named_expressions
        )
    else:
        with pytest.raises(RuntimeError):
            NamedExpressions(settings_source=solver)
    assert Methods(settings_source=solver) == solver.solution.methods
    assert Controls(settings_source=solver) == solver.solution.controls
    assert (
        ReportDefinitions(settings_source=solver) == solver.solution.report_definitions
    )
    if fluent_version >= FluentVersion.v231:
        assert Monitor(settings_source=solver) == solver.solution.monitor
        if fluent_version >= FluentVersion.v241:
            assert Residual(settings_source=solver) == solver.solution.monitor.residual
        else:
            with pytest.raises(RuntimeError):
                Residual(settings_source=solver)
        assert (
            ReportFiles(settings_source=solver) == solver.solution.monitor.report_files
        )
        assert (
            ReportFile(settings_source=solver, new_instance_name="report-file-1")
            == solver.solution.monitor.report_files["report-file-1"]
        )
        assert (
            ReportFile(settings_source=solver, name="report-file-1")
            == solver.solution.monitor.report_files["report-file-1"]
        )
        if fluent_version >= FluentVersion.v251:
            assert (
                ReportFile(settings_source=solver)
                == solver.solution.monitor.report_files["report-file-2"]
            )
        assert (
            ReportPlots(settings_source=solver) == solver.solution.monitor.report_plots
        )
        assert (
            ReportPlot(settings_source=solver, new_instance_name="report-plot-1")
            == solver.solution.monitor.report_plots["report-plot-1"]
        )
        assert (
            ReportPlot(settings_source=solver, name="report-plot-1")
            == solver.solution.monitor.report_plots["report-plot-1"]
        )
        if fluent_version >= FluentVersion.v251:
            assert (
                ReportPlot(settings_source=solver)
                == solver.solution.monitor.report_plots["report-plot-2"]
            )
        assert (
            ConvergenceConditions(settings_source=solver)
            == solver.solution.monitor.convergence_conditions
        )
    else:
        with pytest.raises(RuntimeError):
            Monitor(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert CellRegisters(settings_source=solver) == solver.solution.cell_registers
        assert (
            CellRegister(settings_source=solver, new_instance_name="cell_register_1")
            == solver.solution.cell_registers["cell_register_1"]
        )
        assert (
            CellRegister(settings_source=solver, name="cell_register_1")
            == solver.solution.cell_registers["cell_register_1"]
        )
        if fluent_version >= FluentVersion.v251:
            assert (
                CellRegister(settings_source=solver)
                == solver.solution.cell_registers["cell_register_2"]
            )
    else:
        with pytest.raises(RuntimeError):
            CellRegisters(settings_source=solver)
    assert Initialization(settings_source=solver) == solver.solution.initialization
    if fluent_version >= FluentVersion.v231:
        assert (
            CalculationActivity(settings_source=solver)
            == solver.solution.calculation_activity
        )
        assert (
            ExecuteCommands(settings_source=solver)
            == solver.solution.calculation_activity.execute_commands
        )
        if fluent_version >= FluentVersion.v241:
            assert (
                CaseModification(settings_source=solver)
                == solver.solution.calculation_activity.case_modification
            )
        else:
            with pytest.raises(RuntimeError):
                CaseModification(settings_source=solver)
    else:
        with pytest.raises(RuntimeError):
            CalculationActivity(settings_source=solver)
    assert RunCalculation(settings_source=solver) == solver.solution.run_calculation
    assert Solution(settings_source=solver) == solver.solution
    assert Results(settings_source=solver) == solver.results
    assert Surfaces(settings_source=solver) == solver.results.surfaces
    if fluent_version >= FluentVersion.v232:
        assert (
            PointSurfaces(settings_source=solver)
            == solver.results.surfaces.point_surface
        )
        assert (
            LineSurfaces(settings_source=solver) == solver.results.surfaces.line_surface
        )
        assert (
            RakeSurfaces(settings_source=solver) == solver.results.surfaces.rake_surface
        )
        assert (
            IsoSurfaces(settings_source=solver) == solver.results.surfaces.iso_surface
        )
    else:
        with pytest.raises(RuntimeError):
            PointSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            LineSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            RakeSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            IsoSurfaces(settings_source=solver)
    assert (
        PlaneSurfaces(settings_source=solver) == solver.results.surfaces.plane_surface
    )
    assert (
        PlaneSurface(settings_source=solver, new_instance_name="plane-1")
        == solver.results.surfaces.plane_surface["plane-1"]
    )
    assert (
        PlaneSurface(settings_source=solver, name="plane-1")
        == solver.results.surfaces.plane_surface["plane-1"]
    )
    if fluent_version >= FluentVersion.v241:
        assert IsoClips(settings_source=solver) == solver.results.surfaces.iso_clip
    else:
        with pytest.raises(RuntimeError):
            IsoClips(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert (
            ZoneSurfaces(settings_source=solver) == solver.results.surfaces.zone_surface
        )
        assert (
            PartitionSurfaces(settings_source=solver)
            == solver.results.surfaces.partition_surface
        )
        assert (
            TransformSurfaces(settings_source=solver)
            == solver.results.surfaces.transform_surface
        )
        assert (
            ImprintSurfaces(settings_source=solver)
            == solver.results.surfaces.imprint_surface
        )
        assert (
            PlaneSlices(settings_source=solver) == solver.results.surfaces.plane_slice
        )
        assert (
            SphereSlices(settings_source=solver) == solver.results.surfaces.sphere_slice
        )
        assert (
            QuadricSurfaces(settings_source=solver)
            == solver.results.surfaces.quadric_surface
        )
        assert (
            SurfaceCells(settings_source=solver)
            == solver.results.surfaces.surface_cells
        )
    else:
        with pytest.raises(RuntimeError):
            ZoneSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            PartitionSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            TransformSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            ImprintSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            PlaneSlices(settings_source=solver)
        with pytest.raises(RuntimeError):
            SphereSlices(settings_source=solver)
        with pytest.raises(RuntimeError):
            QuadricSurfaces(settings_source=solver)
        with pytest.raises(RuntimeError):
            SurfaceCells(settings_source=solver)
    if fluent_version >= FluentVersion.v251:
        assert (
            ExpressionVolumes(settings_source=solver)
            == solver.results.surfaces.expression_volume
        )
        assert (
            GroupSurfaces(settings_source=solver)
            == solver.results.surfaces.group_surface
        )
    else:
        with pytest.raises(RuntimeError):
            ExpressionVolumes(settings_source=solver)
        with pytest.raises(RuntimeError):
            GroupSurfaces(settings_source=solver)
    assert Graphics(settings_source=solver) == solver.results.graphics
    assert Meshes(settings_source=solver) == solver.results.graphics.mesh
    assert Contours(settings_source=solver) == solver.results.graphics.contour
    assert (
        Contour(settings_source=solver, new_instance_name="contour-1")
        == solver.results.graphics.contour["contour-1"]
    )
    assert (
        Contour(settings_source=solver, name="contour-1")
        == solver.results.graphics.contour["contour-1"]
    )
    assert Vectors(settings_source=solver) == solver.results.graphics.vector
    assert LICs(settings_source=solver) == solver.results.graphics.lic
    if fluent_version >= FluentVersion.v231:
        assert Pathlines(settings_source=solver) == solver.results.graphics.pathline
        assert (
            ParticleTracks(settings_source=solver)
            == solver.results.graphics.particle_track
        )
        assert Plots(settings_source=solver) == solver.results.plot
        assert XYPlots(settings_source=solver) == solver.results.plot.xy_plot
    else:
        with pytest.raises(RuntimeError):
            Pathlines(settings_source=solver)
        with pytest.raises(RuntimeError):
            ParticleTracks(settings_source=solver)
        with pytest.raises(RuntimeError):
            Plots(settings_source=solver)
        with pytest.raises(RuntimeError):
            XYPlots(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert Histogram(settings_source=solver) == solver.results.plot.histogram
        assert (
            CumulativePlots(settings_source=solver)
            == solver.results.plot.cumulative_plot
        )
    else:
        with pytest.raises(RuntimeError):
            Histogram(settings_source=solver)
        with pytest.raises(RuntimeError):
            CumulativePlots(settings_source=solver)
    if fluent_version >= FluentVersion.v242:
        assert ProfileData(settings_source=solver) == solver.results.plot.profile_data
        assert (
            InterpolatedData(settings_source=solver)
            == solver.results.plot.interpolated_data
        )
    else:
        with pytest.raises(RuntimeError):
            ProfileData(settings_source=solver)
        with pytest.raises(RuntimeError):
            InterpolatedData(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert Scenes(settings_source=solver) == solver.results.scene
    else:
        with pytest.raises(RuntimeError):
            Scenes(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert (
            SceneAnimation(settings_source=solver)
            == solver.results.animations.scene_animation
        )
    else:
        with pytest.raises(RuntimeError):
            SceneAnimation(settings_source=solver)
    if fluent_version >= FluentVersion.v231:
        assert Report(settings_source=solver) == solver.results.report
        assert (
            DiscretePhaseHistogram(settings_source=solver)
            == solver.results.report.discrete_phase.histogram
        )
        assert Fluxes(settings_source=solver) == solver.results.report.fluxes
        assert (
            SurfaceIntegrals(settings_source=solver)
            == solver.results.report.surface_integrals
        )
        assert (
            VolumeIntegrals(settings_source=solver)
            == solver.results.report.volume_integrals
        )
        assert (
            SimulationReports(settings_source=solver)
            == solver.results.report.simulation_reports
        )
    else:
        with pytest.raises(RuntimeError):
            Report(settings_source=solver)
        with pytest.raises(RuntimeError):
            DiscretePhaseHistogram(settings_source=solver)
        with pytest.raises(RuntimeError):
            Fluxes(settings_source=solver)
        with pytest.raises(RuntimeError):
            SurfaceIntegrals(settings_source=solver)
        with pytest.raises(RuntimeError):
            VolumeIntegrals(settings_source=solver)
        with pytest.raises(RuntimeError):
            SimulationReports(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert (
            InputParameters(settings_source=solver)
            == solver.parameters.input_parameters
        )
        assert (
            OutputParameters(settings_source=solver)
            == solver.parameters.output_parameters
        )
    else:
        with pytest.raises(RuntimeError):
            InputParameters(settings_source=solver)
        with pytest.raises(RuntimeError):
            OutputParameters(settings_source=solver)
    if fluent_version >= FluentVersion.v251:
        assert (
            CustomFieldFunctions(settings_source=solver)
            == solver.results.custom_field_functions
        )
    else:
        with pytest.raises(RuntimeError):
            CustomFieldFunctions(settings_source=solver)
    if fluent_version >= FluentVersion.v241:
        assert CustomVectors(settings_source=solver) == solver.results.custom_vectors
    else:
        with pytest.raises(RuntimeError):
            CustomVectors(settings_source=solver)
    tmp_save_path = tempfile.mkdtemp(dir=pyfluent.EXAMPLES_PATH)
    project_file = Path(tmp_save_path) / "mixing_elbow_param.flprj"
    solver.settings.parametric_studies.initialize(project_filename=str(project_file))
    assert ParametricStudies(settings_source=solver) == solver.parametric_studies
    assert (
        ParametricStudy(settings_source=solver, name="mixing_elbow-Solve")
        == solver.parametric_studies["mixing_elbow-Solve"]
    )
    assert (
        DesignPoints(settings_source=solver, parametric_studies="mixing_elbow-Solve")
        == solver.parametric_studies["mixing_elbow-Solve"].design_points
    )
    assert (
        DesignPoint(
            settings_source=solver,
            parametric_studies="mixing_elbow-Solve",
            name="Base DP",
        )
        == solver.parametric_studies["mixing_elbow-Solve"].design_points["Base DP"]
    )


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_builtin_singleton_setting_assign_session(
    new_meshing_session, new_solver_session
):
    meshing = new_meshing_session
    solver = new_solver_session

    models = Models()
    assert isinstance(models, Models)
    with pytest.raises(TypeError):
        models.settings_source = meshing
    models.settings_source = solver
    assert models.settings_source == solver.settings
    assert not isinstance(models, Models)
    assert models.path == "setup/models"
    assert not models.is_active()
    case_name = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    solver.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    assert models.is_active()
    assert models == solver.setup.models
    # TODO: Ideally an AttributeError should be raised here from flobject.py
    with pytest.raises(RuntimeError):  # re-assignment is not allowed
        models.settings_source = solver

    models = Models()
    assert isinstance(models, Models)
    models.settings_source = solver.settings
    assert models == solver.setup.models
    assert models.settings_source == solver.settings


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_builtin_non_creatable_named_object_setting_assign_session(
    new_meshing_session, static_mixer_case_session
):
    meshing = new_meshing_session
    solver = static_mixer_case_session

    inlet = BoundaryCondition(name="inlet1")
    assert isinstance(inlet, BoundaryCondition)
    with pytest.raises(TypeError):
        inlet.settings_source = meshing
    inlet.settings_source = solver
    assert inlet == solver.settings.setup.boundary_conditions["inlet1"]
    assert inlet.settings_source == solver.settings
    # TODO: Ideally an AttributeError should be raised here from flobject.py
    with pytest.raises(RuntimeError):  # re-assignment is not allowed
        inlet.settings_source = solver

    inlet = BoundaryCondition(name="inlet1")
    assert isinstance(inlet, BoundaryCondition)
    inlet.settings_source = solver.settings
    assert inlet == solver.settings.setup.boundary_conditions["inlet1"]
    assert inlet.settings_source == solver.settings


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_builtin_creatable_named_object_setting_assign_session(
    new_meshing_session, static_mixer_case_session
):
    meshing = new_meshing_session
    solver = static_mixer_case_session

    report_file = ReportFile(new_instance_name="report-file-1")
    assert isinstance(report_file, ReportFile)
    with pytest.raises(TypeError):
        report_file.settings_source = meshing
    report_file.settings_source = solver
    assert report_file == solver.solution.monitor.report_files["report-file-1"]
    assert report_file.settings_source == solver.settings
    # TODO: Ideally an AttributeError should be raised here from flobject.py
    with pytest.raises(RuntimeError):  # re-assignment is not allowed
        report_file.settings_source = solver

    report_file = ReportFile(name="report-file-1")
    assert isinstance(report_file, ReportFile)
    report_file.settings_source = solver
    assert report_file == solver.solution.monitor.report_files["report-file-1"]
    assert report_file.settings_source == solver.settings

    report_file = ReportFile(name="report-file-1")
    assert isinstance(report_file, ReportFile)
    report_file.settings_source = solver.settings
    assert report_file == solver.solution.monitor.report_files["report-file-1"]
    assert report_file.settings_source == solver.settings

    if solver.get_fluent_version() >= FluentVersion.v251:
        report_file = ReportFile()
        assert isinstance(report_file, ReportFile)
        report_file.settings_source = solver
        assert report_file == solver.solution.monitor.report_files["report-file-2"]
        assert report_file.settings_source == solver.settings



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
from tempfile import TemporaryDirectory
import uuid

from pytest import MonkeyPatch

import ansys.fluent.core as pyfluent
from ansys.fluent.core.codegen import StaticInfoType, datamodelgen
from ansys.fluent.core.utils import load_module


def create_datamodel_root_in_server(session, rules_str, app_name) -> None:
    rules_file_name = f"{uuid.uuid4()}.fdl"
    session.scheme_eval.scheme_eval(
        f'(with-output-to-file "{rules_file_name}" (lambda () (format "~a" "{rules_str}")))',
    )
    session.scheme_eval.scheme_eval(
        f'(state/register-new-state-engine "{app_name}" "{rules_file_name}")'
    )
    session.scheme_eval.scheme_eval(f'(remove-file "{rules_file_name}")')
    assert session.scheme_eval.scheme_eval(f'(state/find-root "{app_name}")') > 0


def create_root_using_datamodelgen(service, app_name):
    version = "252"
    static_info = service.get_static_info(app_name)
    with TemporaryDirectory() as temp_dir:
        with MonkeyPatch.context() as m:
            m.setattr(pyfluent, "CODEGEN_OUTDIR", Path(temp_dir))
            # TODO: Refactor datamdodelgen so we don't need to hardcode StaticInfoType
            datamodelgen.generate(
                version, static_infos={StaticInfoType.DATAMODEL_WORKFLOW: static_info}
            )
            gen_file = Path(temp_dir) / f"datamodel_{version}" / "workflow.py"
            module = load_module("datamodel", gen_file)
            return module.Root(service, app_name, [])



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

import pytest

from ansys.fluent.core import MeshingEvent
from ansys.fluent.core.examples.downloads import download_file


@pytest.mark.nightly
@pytest.mark.fluent_version("latest")
@pytest.mark.codegen_required
def test_launch_pure_meshing(mixing_elbow_watertight_pure_meshing_session):
    pure_meshing_session = mixing_elbow_watertight_pure_meshing_session
    assert pure_meshing_session.health_check.is_serving
    file_name = "launch_pure_meshing_journal.py"
    pure_meshing_session.journal.start(file_name)
    session_dir = dir(pure_meshing_session)
    for attr in ("fields", "meshing", "workflow"):
        assert attr in session_dir
    workflow = pure_meshing_session.workflow
    workflow.TaskObject["Import Geometry"].Execute()

    add_local_sizing = workflow.TaskObject["Add Local Sizing"]
    add_local_sizing.AddChildToTask()
    add_local_sizing.Execute()

    surface_mesh_gen = workflow.TaskObject["Generate the Surface Mesh"]
    surface_mesh_gen.Arguments = {"CFDSurfaceMeshControls": {"MaxSize": 0.3}}
    surface_mesh_gen.Execute()

    describe_geo = workflow.TaskObject["Describe Geometry"]
    describe_geo.UpdateChildTasks(SetupTypeChanged=False)
    describe_geo.Arguments = dict(
        SetupType="The geometry consists of only fluid regions with no voids"
    )
    describe_geo.UpdateChildTasks(SetupTypeChanged=True)
    describe_geo.Execute()

    boundary_update = workflow.TaskObject["Update Boundaries"]
    boundary_update.Arguments = {
        "BoundaryLabelList": ["wall-inlet"],
        "BoundaryLabelTypeList": ["wall"],
        "OldBoundaryLabelList": ["wall-inlet"],
        "OldBoundaryLabelTypeList": ["velocity-inlet"],
    }
    boundary_update.Execute()

    workflow.TaskObject["Update Regions"].Execute()

    add_boundary_layers = workflow.TaskObject["Add Boundary Layers"]
    add_boundary_layers.AddChildToTask()
    add_boundary_layers.InsertCompoundChildTask()
    smooth_transition_1 = workflow.TaskObject["smooth-transition_1"]
    smooth_transition_1.Arguments = {
        "BLControlName": "smooth-transition_1",
    }
    add_boundary_layers.Arguments = {}
    smooth_transition_1.Execute()
    volume_mesh_gen = workflow.TaskObject["Generate the Volume Mesh"]
    volume_mesh_gen.Arguments = {
        "VolumeFill": "poly-hexcore",
        "VolumeFillControls": {
            "HexMaxCellLength": 0.3,
        },
    }
    volume_mesh_gen.Execute()

    pure_meshing_session.journal.stop()
    with pytest.raises(AttributeError):
        pure_meshing_session.switch_to_solver()
    pure_meshing_session.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    geom_name = download_file("mixing_elbow.pmdb", "pyfluent/mixing_elbow")
    pure_meshing_session.workflow.TaskObject["Import Geometry"].Arguments = dict(
        FileName=geom_name, LengthUnit="in"
    )
    pure_meshing_session.tui.file.read_journal(file_name)
    if os.path.exists(file_name):
        os.remove(file_name)


@pytest.mark.fluent_version("latest")
@pytest.mark.codegen_required
def test_launch_meshing_and_switch(new_meshing_session):
    meshing = new_meshing_session
    assert not meshing._switched
    _ = meshing.switch_to_solver()
    assert meshing._switched
    assert not meshing.tui
    assert not meshing.meshing
    assert not meshing.workflow
    assert not meshing.watertight


@pytest.mark.fluent_version("latest")
@pytest.mark.codegen_required
def test_meshing_streaming_and_switch(new_meshing_session):

    def on_case_loaded(session, event_info):
        on_case_loaded.loaded = True

    on_case_loaded.loaded = False

    def on_trancript(transcript):
        on_trancript.called = True

    on_trancript.called = False

    meshing = new_meshing_session

    meshing.events.register_callback(MeshingEvent.CASE_LOADED, on_case_loaded)
    meshing.transcript.register_callback(on_trancript)

    solver = meshing.switch_to_solver()

    on_trancript.called = False

    case_file_name = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")

    try:
        solver.settings.file.read_case(file_name=case_file_name)
    except AttributeError:
        solver.tui.file.read_case(case_file_name)

    assert not on_trancript.called
    assert not on_case_loaded.loaded


@pytest.mark.fluent_version("latest")
@pytest.mark.codegen_required
def test_meshing_no_foreign_objects(new_meshing_session):
    meshing = new_meshing_session
    with pytest.raises(AttributeError):
        meshing.monitors


def test_fake_session():

    class fake_session_base:

        def bar(self):
            pass

    class fake_session(fake_session_base):

        def __init__(self) -> None:
            self._switched = False

        def __getattribute__(self, item: str):
            if item == "_switched":
                return super(fake_session, self).__getattribute__(item)

            if self._switched:
                return None

            return super(fake_session, self).__getattribute__(item)

        def foo(self):
            return 42

    f = fake_session()

    assert f.foo() == 42

    f._switched = True

    assert f.foo is None

    assert f.bar is None



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _modeling_ablation:

Modeling Ablation
-------------------------------------------
"""

#######################################################################################
# Objective
# =====================================================================================
#
# Ablation is an effective treatment used to protect an atmospheric reentry vehicle from
# the damaging effects of external high temperatures caused by shock wave and viscous
# heating. The ablative material is chipped away due to surface reactions that remove a
# significant amount of heat and keep the vehicle surface temperature below the melting
# point. In this tutorial, Fluent ablation model is demonstrated for a reendtry vehicle
# geometry simplified as a 3D wedge.
#
# This tutorial demonstrates how to do the following:
#
# * Define boundary conditions for a high-speed flow.
# * Set up the ablation model to model effects of a moving boundary due to ablation.
# * Initiate and solve the transient simulation using the density-based solver.
#
# Problem Description:
# ====================
#
# The geometry of the 3D wedge considered in this tutorial is shown in following figure.
# The air flow passes around a nose of a re-entry vehicle operating under high speed
# conditions. The inlet air has a temperature of 4500 K, a gauge pressure of 13500 Pa,
# and a Mach number of 3. The domain is bounded above and below by symmetry planes
# (displayed in yellow). As the ablative coating chips away, the surface of the wall
# moves. The moving of the surface is modeled using dynamic meshes. The surface moving
# rate is estimated by Vieille's empirical law:
#
# where r is the surface moving rate, p is the absolute pressure, and A and n are model
# parameters. In the considered case, A = 5 and n = 0.1.


####################################################################################
# .. math::
#
#    r = A \cdot p^n


# %%
# .. image:: ../../_static/ablation-problem-schematic.png
#    :align: center
#    :alt: Problem Schematic

# %%

####################################################################################
# Import required libraries/modules
# ==================================================================================

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.visualization.pyvista import Graphics

####################################################################################
# Download example file
# ==================================================================================
import_filename = examples.download_file(
    "ablation.msh.h5",
    "pyfluent/examples/Ablation-tutorial",
)

####################################################################################
# Fluent Solution Setup
# ==================================================================================

from ansys.fluent.visualization import set_config  # noqa: E402

set_config(blocking=True, set_view_on_display="isometric")

####################################################################################
# Launch Fluent session with solver mode and print Fluent version
# ==================================================================================

solver = pyfluent.launch_fluent(
    product_version="25.1.0",
    dimension=3,
    precision="double",
    processor_count=4,
)
print(solver.get_fluent_version())

####################################################################################
# Import mesh
# ==================================================================================

solver.file.read_case(file_name=import_filename)

####################################################################################
# Define models
# ==================================================================================

solver.setup.general.solver.type = "density-based-implicit"
solver.setup.general.solver.time = "unsteady-1st-order"
solver.setup.general.operating_conditions.operating_pressure = 0.0
solver.setup.models.energy = {"enabled": True}
solver.setup.models.ablation = {"enabled": True}

###################################################################
# Define material
# =================================================================

solver.setup.materials.fluid["air"] = {
    "density": {"option": "ideal-gas"},
    "molecular_weight": {"value": 28.966, "option": "constant"},
}
solver.setup.materials.fluid["air"] = {"density": {"option": "ideal-gas"}}

############################
# Define boundary conditions
# ==========================

solver.setup.boundary_conditions.set_zone_type(
    zone_list=["inlet"], new_type="pressure-far-field"
)
solver.setup.boundary_conditions.pressure_far_field["inlet"].momentum.gauge_pressure = (
    13500
)
solver.setup.boundary_conditions.pressure_far_field["inlet"].momentum.mach_number = 3
solver.setup.boundary_conditions.pressure_far_field["inlet"].thermal.temperature = 4500
solver.setup.boundary_conditions.pressure_far_field[
    "inlet"
].turbulence.turbulent_intensity = 0.001

solver.setup.boundary_conditions.pressure_outlet["outlet"].momentum.gauge_pressure = (
    13500
)
solver.setup.boundary_conditions.pressure_outlet[
    "outlet"
].momentum.prevent_reverse_flow = True

#############################################
# Ablation boundary condition (Vielles Model)
# ++++++++++++++++++++++++++++++++++++++++++++
# Once you have specified the ablation boundary conditions for the wall,
# Ansys Fluent automatically enables the Dynamic Mesh model with the Smoothing and
# Remeshing options, #creates the wall-ablation dynamic mesh zone, and configure
# appropriate dynamic mesh settings.

solver.setup.boundary_conditions.wall[
    "wall_ablation"
].ablation.ablation_select_model = "Vielle's Model"
solver.setup.boundary_conditions.wall["wall_ablation"].ablation.ablation_vielle_a = 5
solver.setup.boundary_conditions.wall["wall_ablation"].ablation.ablation_vielle_n = 0.1

##############################
# Define dynamic mesh controls
# ============================

solver.tui.define.dynamic_mesh.dynamic_mesh("yes")
solver.tui.define.dynamic_mesh.zones.create(
    "interior--flow",
    "deforming",
    "faceted",
    "no",
    "no",
    "yes",
    "no",
    "yes",
    "yes",
    "no",
    "yes",
)
solver.tui.define.dynamic_mesh.zones.create(
    "outlet",
    "deforming",
    "faceted",
    "no",
    "yes",
    "no",
    "yes",
    "yes",
    "coefficient-based",
    "0.1",
    "yes",
)
solver.tui.define.dynamic_mesh.zones.create(
    "symm1",
    "deforming",
    "plane",
    "0",
    "-0.04",
    "0",
    "0",
    "-1",
    "0",
    "no",
    "yes",
    "no",
    "yes",
    "yes",
    "coefficient-based",
    "0.1",
    "yes",
)
solver.tui.define.dynamic_mesh.zones.create(
    "symm2",
    "deforming",
    "plane",
    "0",
    "0.04",
    "0",
    "0",
    "1",
    "0",
    "no",
    "yes",
    "no",
    "yes",
    "yes",
    "coefficient-based",
    "0.1",
    "yes",
)
solver.tui.define.dynamic_mesh.zones.create(
    "wall_ablation",
    "user-defined",
    "**ablation**",
    "no",
    "no",
    "189",
    "constant",
    "0",
    "yes",
    "yes",
    "0.7",
    "no",
    "no",
)

############################################
# Define solver settings
# =======================

solver.setup.general.solver.time = "unsteady-2nd-order"
solver.solution.controls.limits = {"max_temperature": 25000}
solver.solution.monitor.residual.equations["energy"].absolute_criteria = 1e-06

############################################
# Create report definitions
# ==========================

solver.solution.report_definitions.drag["drag_force_x"] = {}
solver.solution.report_definitions.drag["drag_force_x"].zones = ["wall_ablation"]

solver.solution.monitor.report_plots.create(name="drag_force_x")
solver.solution.monitor.report_plots["drag_force_x"].report_defs = "drag_force_x"
solver.tui.solve.report_plots.axes(
    "drag_force_x", "numbers", "float", "4", "exponential", "2", "q"
)

solver.solution.monitor.report_files.create(name="drag_force_x")
solver.solution.monitor.report_files["drag_force_x"] = {
    "report_defs": ["drag_force_x"],
    "file_name": r"drag_force_x.out",
}

solver.solution.report_definitions.surface["pressure_avg_abl_wall"] = {}
solver.solution.report_definitions.surface["pressure_avg_abl_wall"].report_type = (
    "surface-areaavg"
)
solver.solution.report_definitions.surface["pressure_avg_abl_wall"].field = "pressure"
solver.solution.report_definitions.surface["pressure_avg_abl_wall"].surface_names = [
    "wall_ablation"
]

solver.solution.monitor.report_plots.create(name="pressure_avg_abl_wall")
solver.solution.monitor.report_plots["pressure_avg_abl_wall"].report_defs = (
    "pressure_avg_abl_wall"
)
solver.tui.solve.report_plots.axes(
    "pressure_avg_abl_wall", "numbers", "float", "4", "exponential", "2", "q"
)

solver.solution.monitor.report_files.create(name="pressure_avg_abl_wall")
solver.solution.monitor.report_files["pressure_avg_abl_wall"] = {
    "report_defs": ["pressure_avg_abl_wall"],
    "file_name": r"pressure_avg_abl_wall.out",
}

solver.solution.report_definitions.surface["recede_point"] = {}
solver.solution.report_definitions.surface["recede_point"].report_type = (
    "surface-vertexmax"
)
solver.solution.report_definitions.surface["recede_point"].field = "z-coordinate"
solver.solution.report_definitions.surface["recede_point"].surface_names = [
    "wall_ablation"
]

solver.solution.monitor.report_plots.create(name="recede_point")
solver.solution.monitor.report_plots["recede_point"].report_defs = "recede_point"
solver.tui.solve.report_plots.axes(
    "recede_point", "numbers", "float", "4", "exponential", "2", "q"
)
solver.solution.monitor.report_files.create(name="recede_point")
solver.solution.monitor.report_files["recede_point"] = {
    "report_defs": ["recede_point"],
    "file_name": r"recede_point.out",
}

############################################
# Initialize and Save case
# ========================

solver.tui.solve.initialize.compute_defaults.pressure_far_field("inlet")
solver.solution.initialization.initialization_type = "standard"
solver.solution.initialization.standard_initialize()
solver.solution.run_calculation.transient_controls.time_step_size = 1e-06

solver.file.write(file_type="case", file_name="ablation.cas.h5")

############################################
# Run the calculation
# ===================
# Note: It may take about half an hour to finish the calculation.

solver.solution.run_calculation.dual_time_iterate(
    time_step_count=100, max_iter_per_step=20
)

###############################################
# Save simulation data
# ====================
# Write case and data files
solver.file.write(file_type="case-data", file_name="ablation_Solved.cas.h5")

####################################################################################
# Post Processing
# ==================================================================================

###############################################
# Display plots
# =============

# %%
# .. image:: ../../_static/ablation-residual.png
#    :align: center
#    :alt: residual

# %%
#    Scaled residual plot

# %%
# .. image:: ../../_static/ablation-drag_force_x.png
#    :align: center
#    :alt: Drag force in x direction

# %%
#    History of the drag force on the ablation wall

# %%
# .. image:: ../../_static/ablation-avg_pressure.png
#    :align: center
#    :alt: Average pressure on the ablation wall

# %%
#    History of the averaged pressure on the ablation wall

# %%
# .. image:: ../../_static/ablation-recede_point.png
#    :align: center
#    :alt: Recede point (albation)

# %%
#    Recede point (deformation due to ablation)

###############################################
# Display contour
# ================
# Following contours are displayed in the Fluent GUI:

solver.results.surfaces.plane_surface.create(name="mid_plane")
solver.results.surfaces.plane_surface["mid_plane"].method = "zx-plane"

solver.results.graphics.contour.create(name="contour_pressure")
solver.results.graphics.contour["contour_pressure"] = {
    "field": "pressure",
    "surfaces_list": ["mid_plane"],
}
solver.results.graphics.contour.display(object_name="contour_pressure")

solver.results.graphics.contour.create(name="contour_mach")
solver.results.graphics.contour["contour_mach"] = {
    "field": "mach-number",
    "surfaces_list": ["mid_plane"],
}
solver.results.graphics.contour.display(object_name="contour_mach")

###############################################
# Post processing with PyVista (3D visualization)
# ===============================================
# Following graphics is displayed in the a new window/notebook

graphics_session1 = Graphics(solver)
contour1 = graphics_session1.Contours["contour-1"]
contour1.field = "pressure"
contour1.surfaces_list = ["mid_plane"]
contour1.display()
# %%
# .. image:: ../../_static/ablation-pressure.png
#    :align: center
#    :alt: Static Pressure Contour

# %%
#    Static Pressure Contour

contour1.field = "mach-number"
contour1.range.option = "auto-range-off"
contour1.range.auto_range_off.minimum = 0.5
contour1.range.auto_range_off.maximum = 3.0
contour1.display()

# %%
# .. image:: ../../_static/ablation-mach-number.png
#    :align: center
#    :alt: Mach Number Contour

# %%
#    Mach Number Contour

####################################################################################
# Close the session
# ==================================================================================

solver.exit()

# sphinx_gallery_thumbnail_path = '_static/ablation-mach-number-thumbnail.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""ansys-tools-report."""

import ansys.tools.report as pyansys_report

ANSYS_ENV_VARS = [
    "ANSYSLMD_LICENSE_FILE",
    "AWP_ROOT<NNN>",
    "FLUENT_CONTAINER_IMAGE",
    "FLUENT_IMAGE_NAME",
    "FLUENT_IMAGE_TAG",
    "PYFLUENT_CODEGEN_OUTDIR",
    "PYFLUENT_CODEGEN_SKIP_BUILTIN_SETTINGS",
    "PYFLUENT_CONTAINER_MOUNT_SOURCE",
    "PYFLUENT_CONTAINER_MOUNT_TARGET",
    "PYFLUENT_FLUENT_DEBUG",
    "PYFLUENT_DOC_SKIP_CHEATSHEET:",
    "PYFLUENT_DOC_SKIP_EXAMPLES",
    "PYFLUENT_FLUENT_IP",
    "PYFLUENT_FLUENT_PORT",
    "PYFLUENT_FLUENT_ROOT",
    "PYFLUENT_GRPC_LOG_BYTES_LIMIT",
    "PYFLUENT_LAUNCH_CONTAINER",
    "PYFLUENT_LOGGING",
    "PYFLUENT_NO_FIX_PARAMETER_LIST_RETURN",
    "PYFLUENT_SHOW_SERVER_GUI",
    "PYFLUENT_SKIP_API_UPGRADE_ADVICE",
    "PYFLUENT_TIMEOUT_FORCE_EXIT",
    "PYFLUENT_WATCHDOG_DEBUG",
    "PYFLUENT_WATCHDOG_EXCEPTION_ON_ERROR",
    "REMOTING_PORTS",
    "REMOTING_SERVER_ADDRESS",
    "SERVER_INFO_DIR",
]

dependencies = {
    "ansys-api-fluent": "0.3.34",
    "ansys-platform-instancemanagement": "1.1",
    "ansys-tools-filetransfer": "0.1,<0.3",
    "ansys-units": "0.3.3,<0.5",
    "docker": "7.1.0",
    "grpcio": "1.30.0",
    "grpcio-health-checking": "1.30.0",
    "grpcio-status": "1.30.0",
    "h5py": "3.12.1",
    "nltk": "3.9.1",
    "numpy": "1.14.0,<3.0.0",
    "pandas": "1.1.0,<2.3",
    "pyansys-tools-report": "0.8.1",
    "pyyaml": "6.0",
}

if __name__ == "__main__":
    rep = pyansys_report.Report(ansys_libs=dependencies, ansys_vars=ANSYS_ENV_VARS)
    print(rep)
